import os
import re

import geopandas as gpd
import numpy as np
import pandas as pd
from osgeo import ogr
from shapely import wkb
from shapely.geometry import LineString, MultiPolygon, Point, Polygon
import logging

from copy import deepcopy

from delft3dfmpy.core import geometry

class ExtendedGeoDataFrame(gpd.GeoDataFrame):

    # normal properties
    _metadata = ['required_columns', 'geotype'] + gpd.GeoDataFrame._metadata

    def __init__(self, geotype, required_columns=None,logger=logging, *args, **kwargs):

        # Add logger object to self
        # FIXME: logging below results in error
        #self.logger = logger

        # Check type
        if required_columns is None:
            required_columns = []
        elif not isinstance(required_columns, list):
            required_columns = [required_columns]

        # Add required columns to column list
        if 'columns' in kwargs.keys():
            kwargs['columns'] += required_columns
        else:
            kwargs['columns'] = required_columns

        super(ExtendedGeoDataFrame, self).__init__(*args, **kwargs)

        self.required_columns = required_columns[:]
        self.geotype = geotype

    # def drop(self,edf, item, index_col=None,axis=None):
    #     #edf = ExtendedGeoDataFrame(geotype=LineString)
    #     temp = gpd.GeoDataFrame()
    #     for field in self.iteritems():
    #         temp[field[0]] = field[1]
    #     temp.drop(item,axis=axis, inplace=True)
    #     edf.set_data(temp, index_col=index_col, check_columns=True)
    #     return edf

    def copy(self, deep=True):
        """
        Create a copy
        """

        index = self.index.tolist() if not deep else deepcopy(self.index.tolist())
        columns = self.columns.tolist() if not deep else deepcopy(self.columns.tolist())

        edf = ExtendedGeoDataFrame(
            geotype=self.geotype,
            required_columns=[],
            index=index,
            columns=columns,
        )
        edf.required_columns.extend(self.required_columns[:])
        edf.loc[:, :] = self.values if not deep else deepcopy(self.values)

        return edf

    def delete_all(self):
        """
        Empty the dataframe
        """
        if not self.empty:
            self.iloc[:, 0] = np.nan
            self.dropna(inplace=True)

    def read_shp(self, path, index_col=None, column_mapping=None, check_columns=True, proj_crs = None, clip=None, check_geotype=True,
                 id_col='code', filter_cols=False, filter_rows=None, logger=logging):
        """
        Import function, extended with type checks. Does not destroy reference to object.
        """
        # Read GeoDataFrame
        gdf = gpd.read_file(path)

        # Only keep required columns
        if filter_cols:
            logger.info(f'Filtering required column keys')
            gdf.drop(columns=gdf.columns[~gdf.columns.isin(self.required_columns)], inplace=True)

        # filter out rows on key/value pairs if required
        if filter_rows is not None:
            logger.info(f'Filter rows using key value pairs')
            filter = (gdf[list(filter_rows)] == pd.Series(filter_rows)).all(axis=1)
            gdf = gdf[filter]

        # Drop features without geometry
        total_features = len(gdf)
        missing_features = len(gdf.index[gdf.geometry.isnull()])
        gdf.drop(gdf.index[gdf.geometry.isnull()], inplace=True) # temporary fix
        logger.debug(f'{missing_features} out of {total_features} do not have a geometry')

        # Rename columns:
        if column_mapping is not None:
            gdf.rename(columns=column_mapping, inplace=True)

        if 'MultiPolygon' or 'MultiLineString' in str(gdf.geometry.type):
            gdf = gdf.explode()
            for ftc in gdf[id_col].unique():
                if len(gdf[gdf[id_col]==ftc]) > 1:
                    gdf.loc[gdf[id_col]==ftc, id_col] = [f'{i}_{n}' for n, i in enumerate(gdf[gdf[id_col]==ftc][id_col])]
                    logger.info('%s is MultiPolygon; split into single parts.' % ftc)

        # Check number of entries
        if gdf.empty:
            raise IOError('Imported shapefile contains no rows.')

        # Add data to class GeoDataFrame
        self.set_data(gdf, index_col=index_col, check_columns=check_columns, check_geotype=check_geotype)

        # Clip if extent is provided
        if clip is not None:
            self.clip(clip)

        # To re-project CRS system to projected CRS, first all empty geometries should be dropped
        if proj_crs is not None:
            self.check_projection(proj_crs)
        else:
            logger.debug(f'No projected CRS is given in ini-file')

    def set_data(self, gdf, index_col=None, check_columns=True, check_geotype=True):

        if not self.empty:
            self.delete_all()

        # Check columns
        if check_columns:
            self._check_columns(gdf)

        # Copy content
        for col, values in gdf.iteritems():
            self[col] = values.values

        if index_col is None:
            self.index = gdf.index
            self.index.name = gdf.index.name

        else:
            self.index = gdf[index_col]
            self.index.name = index_col

        # Check geometry types
        if check_geotype:
            self._check_geotype()

    def _check_columns(self, gdf):
        """
        Check presence of columns in geodataframe
        """
        present_columns = gdf.columns.tolist()
        for column in self.required_columns:
            if column not in present_columns:
                raise KeyError('Column "{}" not found. Got {}, Expected at least {}'.format(
                    column, ', '.join(present_columns), ', '.join(self.required_columns)))

    def _check_geotype(self):
        """
        Check geometry type
        """
        if not all(isinstance(geo, self.geotype) for geo in self.geometry):
            raise TypeError('Geometrytype "{}" required. The input shapefile has geometry type(s) {}.'.format(
                re.findall('([A-Z].*)\'', repr(self.geotype))[0],
                self.geometry.type.unique().tolist()
            ))

    def read_gml(self, gml_path, index_col=None, groupby_column=None, order_column=None,
                 id_col='code', column_mapping={}, check_columns=True, check_geotype=True, clip=None):
        """
        Read GML file to GeoDataFrame.

        This function has the option to group Points into LineStrings. To do so,
        specify the groupby column (which the set has in common) and the order_column,
        the column which indicates the order of the grouping.

        A mask file can be specified to clip the selection.

        Parameters
        ----------
        gml_path : str
            Path to the GML file
        groupby_column : str
            Optional, column to group points by
        order_column : str
            Optional, columns to specify the order of the grouped points
        mask_file : str
            File containing the mask to clip points.
        """

        if not os.path.exists(gml_path):
            raise OSError(f'File not found: "{gml_path}"')

        ogr.UseExceptions()
        gml = ogr.Open(gml_path)
        layer = gml.GetLayer()
        layerDefinition = layer.GetLayerDefn()
        nfields = layerDefinition.GetFieldCount()

        # Get column names for features
        columns = [layerDefinition.GetFieldDefn(i).GetName() for i in range(nfields)]

        # Collect features
        features = [f for f in layer]

        # Get geometries
        georefs = [f.GetGeometryRef() for f in features]
        for i, geo in enumerate(georefs):
            if geo is None:
                print('Skipping invalid geometry.')
                del georefs[i]
                del features[i]

        geometries = []
        new_feats = []
        for i,f in enumerate(features):
            geometry = wkb.loads(georefs[i].ExportToWkb())
            if(geometry.type=='MultiPolygon'):
                new_geoms = list(geometry)
                geometries.extend(new_geoms)
                new_features = [f]*len(new_geoms)
                new_feats.extend(new_features)
            else:
                geometries.append(geometry)
                new_feats.append(f)
        features = new_feats
        #geometries = [wkb.loads(geo.ExportToWkb()) for geo in georefs]


        # Get group by columns
        if groupby_column is not None:
            # Check if the group by column is found
            if groupby_column not in columns:
                raise ValueError('Groupby column not found in feature list.')

            # Check if the geometry is as expected
            if not isinstance(geometries[0], Point):
                raise ValueError('Can only group Points to LineString')

            # Get the values from the column that is grouped
            columnid = columns.index(groupby_column)
            groupbyvalues = [f.GetField(columnid) for f in features]
            volgid = columns.index(order_column)
            order = [f.GetField(volgid) for f in features]

            # Create empty dict for lines
            branches, counts = np.unique(groupbyvalues, return_counts=True)
            lines = {branch: [0] * count for branch, count in zip(branches, counts)}

            # Since the order does not always start at 1, find the starting number per group
            startnr = {branch: len(features)+1 for branch in branches}
            for branch, volgnr in zip(groupbyvalues, order):
                startnr[branch] = min(volgnr, startnr[branch])

            # Determine relative order of points in profile (required if the point numbering is not subsequent)
            order_rel = []
            for branch, volgnr in zip(groupbyvalues, order):
                lst_volgnr = [x[1] for x in zip(groupbyvalues, order) if x[0] == branch]
                lst_volgnr.sort()
                for i, x in enumerate(lst_volgnr):
                    if volgnr == x:
                        order_rel.append(i)

            # Filter branches with too few points
            singlepoint = (counts < 2)

            # Assign points
            for point, volgnr, branch, volgnr_rel in zip(geometries, order, groupbyvalues, order_rel):
                #lines[branch][volgnr - startnr[branch]] = point
                lines[branch][volgnr_rel] = point

            # Group geometries to lines
            for branch in branches[~singlepoint]:
                if any(isinstance(pt, int) for pt in lines[branch]):
                    print(f'Points are not properly assigned for branch "{branch}". Check the GML.')
                    lines[branch] = [pt for pt in lines[branch] if not isinstance(pt, int)]
                lines[branch] = LineString(lines[branch])

            # Set order for branches with single point to 0, so features are not loaded
            for branch in branches[singlepoint]:
                order[groupbyvalues.index(branch)] = 0

            # Read fields at first occurence
            startnrs = [startnr[branch] for branch in groupbyvalues]
            fields = [list(map(f.GetField, range(nfields))) for i, volgnr, f in zip(order, startnrs, features) if i == volgnr]

            # Get geometries in correct order for featuresF
            geometries = [lines[row[columnid]] for row in fields]

        else:
            fields = [list(map(f.GetField, range(nfields))) for f in features]

        # Create geodataframe
        gdf = gpd.GeoDataFrame(fields, columns=columns, geometry=geometries)
        gdf.rename(columns=column_mapping, inplace=True)

        # add a letter to 'exploded' multipolygons
        #sfx = ['_'+str(i) for i in range(100)]
        for ftc in gdf[id_col].unique():
            if len(gdf[gdf[id_col]==ftc])>1:
                gdf.loc[gdf[id_col]==ftc,id_col] = [f'{i}_{n}' for n, i in enumerate(gdf[gdf[id_col]==ftc][id_col])]
                print(f'{ftc} is MultiPolygon; split into single parts.')

        # Add data to class GeoDataFrame
        self.set_data(gdf, index_col=index_col, check_columns=check_columns, check_geotype=check_geotype)

        if clip is not None:
            self.clip(clip)

    def clip(self, geometry):
        """
        Clip geometry
        """
        if not isinstance(geometry, (Polygon, MultiPolygon)):
            raise TypeError('Expected geometry of type Polygon or MultiPolygon')

        # Clip if needed
        gdf = self.loc[self.intersects(geometry).values]
        if gdf.empty:
            raise ValueError('Found no features within extent geometry.')

        self.set_data(gdf)

    def check_projection(self, crs_out):
        """
        Check if reprojection is required
        """
        if crs_out!=self.crs:
            self.to_crs(crs_out, inplace=True)
        else:
            logger.info(f'OSM data has same projection as projected crs in ini-file')

    def branch_to_prof(self, offset=0., vertex_end=False, rename_col=None, prefix='', suffix=''):
        """Create profiles on branches from branch data"""

        gdf_out = self.copy()

        # interpolate over feature geometries
        if vertex_end:
            chainage = self.length - offset
            p = self.interpolate(chainage)
        else:
            chainage = offset
            p = self.interpolate(chainage)
        gdf_out.geometry = p
        gdf_out['offset'] = chainage

        if rename_col is not None:
            try:
                gdf_out['branch_id'] = gdf_out[rename_col]
                gdf_out[rename_col] = [f'{prefix}{g[1][rename_col]}{suffix}' for g in self.iterrows()]
            except:
                raise ValueError(f"Column rename with '{rename_col}' did not succeed.")

        return gdf_out


    def merge_columns(self, col1, col2, rename_col):
        """merge columns"""

        #if (not(col1 in self) or not(col2 in self)):
        #    raise ValueError(f"'{col1}' or '{col2}' do not exist.")

        if col1 or col2 in self.columns.values:
            try:
                self[rename_col] = self[col1] + self[col2]
            except:
                raise ValueError(f"Merge of two profile columns'{col1}' and '{col2}' did not succeed.")


    def snap_to_branch(self, branches, snap_method, maxdist=5):
        """Snap the geometries to the branch"""
        geometry.find_nearest_branch(branches=branches, geometries=self, method=snap_method, maxdist=maxdist)

class ExtendedDataFrame(pd.DataFrame):

    _metadata = ['required_columns'] + pd.DataFrame._metadata

    def __init__(self, required_columns=None, *args, **kwargs):
        super(ExtendedDataFrame, self).__init__(*args, **kwargs)

        if required_columns is None:
            required_columns = []

        self.required_columns = required_columns[:] if isinstance(required_columns, list) else [required_columns]

    def delete_all(self):
        """
        Empty the dataframe
        """
        if not self.empty:
            self.iloc[:, 0] = np.nan
            self.dropna(inplace=True)

    def set_data(self, df, index_col):

        if not self.empty:
            self.delete_all()

        # Copy content
        for col, values in df.iteritems():
            self[col] = values.values

        if index_col is None:
            self.index = df.index
            self.index.name = df.index.name

        else:
            self.index = df[index_col]
            self.index.name = index_col

        # Check columns and types
        self._check_columns()

    def add_data(self, df):

        if not np.in1d(df.columns, self.columns).all():
            raise KeyError('The new df contains columns that are not present in the current df.')

        # Concatenate data
        current = pd.DataFrame(self.values, index=self.index, columns=self.columns)
        newdf = pd.concat([current, df], ignore_index=False, sort=False)

        # Empty current df
        self.delete_all()

        # Add values again
        self.set_data(newdf, index_col=self.index.name)

    def _check_columns(self):
        """
        Check presence of columns in geodataframe
        """
        present_columns = self.columns.tolist()
        for i, column in enumerate(self.required_columns):
            if column not in present_columns:
                raise KeyError('Column "{}" not found. Got {}, Expected at least {}'.format(
                    column, ', '.join(present_columns), ', '.join(self.required_columns)))


    def read_gml(self, gml_path, index_col=None):
        """
        Read GML file to GeoDataFrame.

        This function has the option to group Points into LineStrings. To do so,
        specify the groupby column (which the set has in common) and the order_column,
        the column which indicates the order of the grouping.

        A mask file can be specified to clip the selection.

        Parameters
        ----------
        gml_path : str
            Path to the GML file
        index_col : str
            Optional, column to be set as index
        """

        if not os.path.exists(gml_path):
            raise OSError(f'File not found: "{gml_path}"')

        ogr.UseExceptions()
        gml = ogr.Open(gml_path)
        layer = gml.GetLayer()
        layer_definition = layer.GetLayerDefn()
        nfields = layer_definition.GetFieldCount()

        # Get column names for features
        columns = [layer_definition.GetFieldDefn(i).GetName() for i in range(nfields)]

        # Collect features
        features = [f for f in layer]

        # Read fields
        fields = [list(map(f.GetField, range(nfields))) for f in features]

        # Create geodataframe
        gmldf = pd.DataFrame(fields, columns=columns)

        # Add data to class GeoDataFrame
        self.set_data(gmldf, index_col=index_col)




import itertools
import logging
import os
   
import geopandas as gpd
import numpy as np
import pandas as pd
import tqdm
from scipy.spatial import KDTree
from shapely.geometry import LineString, Point, Polygon

from delft3dfmpy.converters import hydamo_to_dflowrr
from delft3dfmpy.core import checks, geometry
from delft3dfmpy.datamodels.common import ExtendedGeoDataFrame
from delft3dfmpy.datamodels.cstructures import meshgeom, meshgeomdim
from delft3dfmpy.io import drrreader

logger = logging.getLogger(__name__)

class DFlowRRModel:
    """Main data structure for RR-model in DflowFM. Contains subclasses
    for unpaved, paved,greehouse and open water nodes and external forcings (seepage, precipitation, evaporation)
      """

    def __init__(self):

        
        self.d3b_parameters = {}       

        self.unpaved = Unpaved(self)
        
        self.paved = Paved(self)
        
        self.greenhouse = Greenhouse(self)               
        
        self.openwater = Openwater(self)               
                
        self.external_forcings = ExternalForcings(self)
        
        self.dimr_path = ''
    
class ExternalForcings:
    """
    Class for external forcings, which contains the boundary
    conditions and the initial conditions.
    """

    def __init__(self, dflowrrmodel):
        # Point to relevant attributes from parent
        self.dflowrrmodel = dflowrrmodel
        self.io = drrreader.ExternalForcingsIO(self)
        
        self.boundary_nodes = {}
        self.seepage = {}
        self.precip = {}
        self.evap = {}
        
    def add_precip(self, id, series):
        self.precip[id] = {
            'precip' : series
        }
        
    def add_evap(self, id, series):
        self.evap[id] = {
            'evap' : series
        }
        
    def add_seepage(self, id, series):
        self.seepage[id] = {
            'seepage' : series
        }
    
    def add_boundary_node(self, id, px, py):
        self.boundary_nodes[id] = {
            'id' : id,  
            'px' : px,
            'py' : py            
        }
    
class Unpaved:
    """
    Class for unpaved nodes
    """
        
    def __init__(self, dflowrrmodel):
        # Point to relevant attributes from parent
        self.dflowrrmodel = dflowrrmodel
        
        # initialize a dataframe for every type of nodes related to 'unpaved'
        self.unp_nodes = {}    
        self.ernst_defs = {}
        
        self.io = drrreader.UnpavedIO(self)

    def add_unpaved(self,id, total_area, lu_areas, surface_level, soiltype, surface_storage, infiltration_capacity, initial_gwd, meteo_area, px, py, boundary_node):
        self.unp_nodes[id] = {
            'id' : 'unp_'+id,  
            'na' : '16',
            'ar' : lu_areas,
            'ga' : total_area,
            'lv' : surface_level,
            'co' : '3',
            'su' : '0',
            'sd' : surface_storage,
            'sp' : 'sep_'+id,
            'ic' : infiltration_capacity,
            'ed' : 'ernst_'+id,
            'bt' : soiltype,
            'ig' : initial_gwd,
            'mg' : surface_level,
            'gl' : '1.5',
            'is' : '0',
            'ms' : 'ms_'+meteo_area,
            'px': px,
            'py': py,
            'boundary_node': boundary_node            
        }
        
    def add_ernst_def(self,id, cvo, lv, cvi, cvs):
        self.ernst_defs[id] = {
            'id' : 'ernst_'+id,            
            'cvi' : cvi,
            'cvs' : cvs,
            'cvo' : cvo,
            'lv'  : lv            
        }        
        
class Paved:
    """
    Class for paved nodes.
    """

    def __init__(self, dflowrrmodel):
        # Point to relevant attributes from parent
        self.dflowrrmodel = dflowrrmodel        
        self.pav_nodes = {}
        self.io = drrreader.PavedIO(self)
        
        self.node_geom = {}
        self.link_geom = {}
        
    #PAVE id 'pav_Nde_n003' ar 16200 lv 1 sd '1' ss 0 qc 0 1.94E-05 0 qo 2 2 ms 'Station1' aaf 1 is 0 np 0 dw '1' ro 0 ru 0 qh '' pave#
    def add_paved(self,id,  area, surface_level, street_storage, sewer_storage, pump_capacity, meteo_area, px, py, boundary_node):
        self.pav_nodes[id] = {
            'id' : 'pav_'+id,        
            'ar' : area,
            'lv' : surface_level,            
            'qc' : pump_capacity,            
            'strs' : street_storage,
            'sews' : sewer_storage,            
            'ms' : 'ms_'+meteo_area,
            'is' : '0',
            'np' : '0',
            'ro' : '0',
            'ru' : '0',
            'px': px,
            'py': py,
            'boundary_node': boundary_node
        }                        
                
class Greenhouse:
    """
    Class for greenhouse nodes
    """
    def __init__(self, dflowrrmodel):        
                
        self.dflowrrmodel = dflowrrmodel
        self.gh_nodes = {}
        # Create the io class
        self.io = drrreader.GreenhouseIO(self)

#    GRHS id ’1’ na 10 ar 1000. 0. 0. 3000. 0. 0. 0. 0. 0. 0. sl 1.0 as 0. sd ’roofstor 1mm’ si
#    ’silo typ1’ ms ’meteostat1’ is 50.0 grhs

    def add_greenhouse(self, id, area, surface_level, roof_storage, meteo_area, px, py, boundary_node):
        self.gh_nodes[id] = {            
            'id': 'gh_'+id,
            'ar' : area,
            'sl': surface_level,            
            'sd': roof_storage,
            'ms' : 'ms_'+meteo_area,
            'is' : '0',
            'px': px,
            'py': py,
            'boundary_node': boundary_node
        }

class Openwater:
    """
    Class for open water nodes
    """
    def __init__(self, dflowrrmodel):        
                
        self.dflowrrmodel = dflowrrmodel
        self.ow_nodes = {}
        # Create the io class
        self.io = drrreader.OpenwaterIO(self)

    def add_openwater(self, id, area, meteo_area, px, py, boundary_node):
        self.ow_nodes[id] = {            
            'id': 'ow_'+id,
            'ar' : area,
            'ms' : 'ms_'+meteo_area,
            'px': px,
            'py': py,
            'boundary_node': boundary_node
        }
         


import netCDF4
import numpy as np
from delft3dfmpy.datamodels.cstructures import meshgeom, meshgeomdim

import logging

logger = logging.getLogger(__name__)

ugrid_dim_dict = {
    "network1d_nEdges": ('1d', 'nbranches'),
    "network1d_nNodes": ('1d', 'nnodes'),
    "network1d_nGeometryNodes": ('1d', 'ngeometry'),    
    "mesh1d_nEdges": ('1d', 'numedge'),
    "mesh1d_nNodes": ('1d', 'numnode'),
    "max_nmesh2d_face_nodes": ('2d', 'maxnumfacenodes'),
    "mesh2d_nEdges": ('2d', 'numedge'),
    "mesh2d_nFaces": ('2d', 'numface'),
    "mesh2d_nNodes": ('2d', 'numnode')
}

ugrid_var_dict = {
    'network1d_node_id': ('1d', 'network_node_ids'),
    'network1d_node_long_name': ('1d', 'network_node_long_names'),
    'network1d_node_x': ('1d', 'nnodex'),
    'network1d_node_y': ('1d', 'nnodey'),
    'network1d_branch_id': ('1d', 'network_branch_ids'),
    'network1d_branch_long_name': ('1d', 'network_branch_long_names'),
    'network1d_edge_length': ('1d', 'nbranchlengths'),
    'network1d_branch_order': ('1d', 'nbranchorder'),
    'network1d_edge_nodes': ('1d', 'nedge_nodes'),
    'network1d_geom_x': ('1d', 'ngeopointx'),
    'network1d_geom_y': ('1d', 'ngeopointy'),
    'network1d_geom_node_count': ('1d', 'nbranchgeometrynodes'),
    'mesh1d_node_id': ('1d', 'mesh1d_node_ids'),
    'mesh1d_node_long_name': ('1d', 'mesh1d_node_long_names'),
    'mesh1d_edge_nodes': ('1d', 'edge_nodes'),
    'mesh1d_node_branch': ('1d', 'branchidx'),
    'mesh1d_node_offset': ('1d', 'branchoffsets'),
    'mesh2d_node_x': ('2d', 'nodex'),
    'mesh2d_node_y': ('2d', 'nodey'),
    'mesh2d_node_z': ('2d', 'nodez'),
    'mesh2d_edge_nodes': ('2d', 'edge_nodes'),
    'mesh2d_face_x': ('2d', 'facex'),
    'mesh2d_face_y': ('2d', 'facey'),
    'mesh2d_face_z': ('2d', 'facez'),
    'mesh2d_face_nodes': ('2d', 'face_nodes')
}

class UgridReader:

    def __init__(self, network):

        self.network = network

    def read_ugrid(self, path):

        """
        Read Ugrid from netcdf and return dflowfm cstructure with grid
        """
        ncfile = netCDF4.Dataset(path)

        # Read mesh1d    
        mesh1d = meshgeom(meshgeomdim())
        read_dimensions(mesh1d.meshgeomdim, '1d', ncfile)
        read_values(mesh1d, '1d', ncfile)
        schematised, branches = mesh1d.process_1d_network()
        self.network.mesh1d.add_from_other(mesh1d)

        # Add branches
        for idx, geometry in branches.items():
            self.network.branches.at[idx, 'geometry'] = geometry
        for idx, geometry in schematised.items():
            self.network.schematised.at[idx, 'geometry'] = geometry

        # Read mesh2d
        mesh2d = meshgeom(meshgeomdim())
        read_dimensions(mesh2d.meshgeomdim, '2d', ncfile)
        read_values(mesh2d, '2d', ncfile)
        self.network.mesh2d.add_from_other(mesh2d)

        # Read links1d2d
        links_1dnode, links_2dface = read_links(ncfile)
        self.network.links1d2d.nodes1d.extend(links_1dnode)
        self.network.links1d2d.faces2d.extend(links_2dface)
        
        ncfile.close()

def read_dimensions(meshgeomdim, readdim, ncfile):
    """
    Function to read dimensions from netcdf file
    """

    assert readdim in ['1d', '2d']

    meshgeomdim.dim = int(readdim[0])

    # Read dimensions
    for ncname, (dim, cname) in ugrid_dim_dict.items():
        if readdim != dim:
            continue

        # Check if variable is in nc file
        if ncname not in ncfile.dimensions.keys():
            logger.error(f'Failed to read dimension "{ncname}" from ncfile for {readdim} mesh.')
        
        value = ncfile.dimensions[ncname].size
        logger.info(f'Read dimension "{ncname}" from ncfile for {readdim} mesh.')
        setattr(meshgeomdim, cname, value)

def read_values(meshgeom, readdim, ncfile):
    """
    Function to read values from netcdf file
    """

    assert readdim in ['1d', '2d']

    # Read variables
    for ncname, (dim, cname) in ugrid_var_dict.items():
        if readdim != dim:
            continue

        # Check if variable is in nc file
        if ncname not in ncfile.variables.keys():
            logger.error(f'Failed to read variable "{ncname}" from ncfile for {readdim} mesh.')
        
        # Read values
        values = ncfile.variables[ncname][:]
        logger.info(f'Read variable "{ncname}" with shape {values.shape} from ncfile for {readdim} mesh.')
        
        # Read description variables (strings)
        if cname in meshgeom.description1d.keys():
            meshgeom.description1d[cname] = list(map(str.strip, netCDF4.chartostring(values)))
        # Read numerical values
        else:
            if values.mask.any():
                values[values.mask] = ncfile.variables[ncname]._FillValue
            meshgeom.set_values(cname, np.ravel(values))
            
def read_links(ncfile):
    """
    Function to read 1d 2d links
    """
    var = 'link1d2d'

    if var not in ncfile.variables.keys():
        return [[], []]
    else:
        return ncfile.variables[var][:, :].T.tolist()


import logging

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import LineString
from rasterstats import zonal_stats
from delft3dfmpy.core import checks, geometry
from delft3dfmpy.datamodels.common import ExtendedDataFrame
import rasterio
import warnings
from rasterio.transform import from_origin
import os
import imod
from tqdm.auto import tqdm
import logging

logger = logging.getLogger(__name__)
def generate_unpaved(catchments, landuse, surface_level, soiltype,  surface_storage, infiltration_capacity, initial_gwd, meteo_areas, zonalstats_alltouched=None):    
    """ 
    Combine all data to form a complete UNPAVED definition. ALso the coordinates for the networ topology are included.
    Zonal statistics are applied to land use, to get the areas per type. The classificition described in the notebook is assumed. From the elevation grid, the median value per catchment is assumed, and for soil type the dominant soil type in the catchment is used. Other parameters can be prescribed as as float (spatially uniform) or as a raster name, in which case the mean value per catchment is used.
    
    """
    all_touched=False if zonalstats_alltouched is None else zonalstats_alltouched
               
    # required rasters
    warnings.filterwarnings('ignore')
    lu_rast, lu_affine = read_raster(landuse, static=True)
    lu_counts = zonal_stats(catchments, lu_rast, affine=lu_affine, categorical=True, all_touched=all_touched)   
    
    rast, affine = read_raster(soiltype, static=True)    
    soiltypes = zonal_stats(catchments, soiltype, affine = affine, stats='majority',all_touched=all_touched)
    
    rast, affine = read_raster(surface_level, static=True)
    mean_elev = zonal_stats(catchments, rast, affine=affine, stats="median",all_touched=all_touched)    
        
    # optional rasters
    if isinstance(surface_storage, str):                
        rast,affine = read_raster(surface_storage, static=True)
        sstores = zonal_stats(catchments, rast, affine=affine, stats="mean",all_touched=True)            
    elif isinstance(surface_storage,int):
        surface_storage = float(surface_storage)
    if isinstance(infiltration_capacity, str):        
        rast,affine = read_raster(infiltration_capacity, static=True)
        infcaps = zonal_stats(catchments, rast, affine=affine, stats="mean",all_touched=True)    
    elif isinstance(infiltration_capacity,int):
        infiltration_capacity = float(infiltration_capacity)
    if isinstance(initial_gwd, str):                
        rast,affine = read_raster(initial_gwd, static=True)
        ini_gwds = zonal_stats(catchments, rast, affine=affine, stats="mean", all_touched=True)       
    elif isinstance(initial_gwd,int):
        initial_gwd = float(initial_gwd)
        
    # get raster cellsize    
    px_area = lu_affine[0] * -lu_affine[4]
    
    unpaved_drr = ExtendedDataFrame(required_columns=['code'])
    unpaved_drr.set_data( pd.DataFrame(np.zeros((len(catchments),12)), 
                                       columns=['code','total_area','lu_areas','mvlevel',
                                                'soiltype','surstor','infcap','initial_gwd','meteostat','px','py','boundary'
                                               ], dtype="str"), index_col='code')
    unpaved_drr.index = catchments.code            
    # HyDAMO Crop code; hydamo name, sobek index, sobek name: 
    # 1 aardappelen   3 potatoes
    # 2 graan         5 grain
    # 3 suikerbiet    4 sugarbeet
    # 4 mais          2 corn
    # 5 overige gew. 15 vegetables
    # 6 bloembollen  10 bulbous plants
    # 7 boomgaard     9 orchard
    # 8 gras          1 grass  
    # 9 loofbos      11 dediduous
    # 10 naaldbos    12 conferous
    # 11 natuuur     13 nature
    # 12 braak       14 fallow    
    sobek_indices = [3,5,4,2,15,10,9,1,11,12,13,14]
    for num, cat in enumerate(catchments.itertuples()):    
        # if no rasterdata could be obtained for this catchment, skip it.
        if mean_elev[num]['median'] is None: 
            logger.warning('No rasterdata available for catchment %s' % cat.code)
            continue        
        tm = [m for m in meteo_areas.itertuples() if m.geometry.contains(cat.geometry.centroid)]
        ms = meteo_areas.iloc[0,:][0] if tm==[] else tm[0].code   
        # find corresponding meteo-station
        #ms = [ms for ms in meteo_areas.itertuples() if ms.geometry.contains(cat.geometry.centroid)]
        #ms = ms[0] if ms != [] else meteo_areas.iloc[0,:][0]
        mapping = np.zeros(16, dtype=int)
        for i in range(1,12):
            if i in lu_counts[num]: mapping[sobek_indices[i-1]-1] = lu_counts[num][i]*px_area            
        lu_map = ' '.join(map(str,mapping))        
        elev = mean_elev[num]['median']
        unpaved_drr.at[cat.code, 'code'] = str(cat.code)
        unpaved_drr.at[cat.code, 'total_area'] = f'{cat.geometry.area:.0f}'
        unpaved_drr.at[cat.code, 'lu_areas'] = lu_map
        unpaved_drr.at[cat.code, 'mvlevel'] = f'{elev:.2f}'
        unpaved_drr.at[cat.code, 'soiltype'] =f'{soiltypes[num]["majority"]+100.:.0f}'        
        if isinstance(surface_storage, float):    
            unpaved_drr.at[cat.code, 'surstor'] = f'{surface_storage:.3f}' 
        else:        
            unpaved_drr.at[cat.code, 'surstor'] = f'{sstores[num]["mean"]:.3f}'
        if isinstance(infiltration_capacity, float):    
            unpaved_drr.at[cat.code, 'infcap'] = f'{infiltration_capacity:.3f}' 
        else:        
            unpaved_drr.at[cat.code, 'infcap'] = f'{infcaps[num]["mean"]:.3f}'
        if isinstance(initial_gwd, float):    
            unpaved_drr.at[cat.code, 'initial_gwd'] = f'{initial_gwd:.2f}'  
        else:
            unpaved_drr.at[cat.code, 'initial_gwd'] = f'{ini_gwds[num]["mean"]:.2f}'
        unpaved_drr.at[cat.code, 'meteostat'] = ms
        unpaved_drr.at[cat.code, 'px'] = f'{cat.geometry.centroid.coords[0][0]-10:.0f}'
        unpaved_drr.at[cat.code, 'py'] = f'{cat.geometry.centroid.coords[0][1]:.0f}'
        unpaved_drr.at[cat.code, 'boundary'] = cat.lateraleknoopcode                
    return unpaved_drr          
    
def generate_ernst(catchments, depths, resistance, infiltration_resistance, runoff_resistance):    
    """
    The lists with depths and resistances as well as the standard infiltration and runoff resistances, are converted, to a dataframe.
    """
    ernst_drr = ExtendedDataFrame(required_columns=['code'])
    ernst_drr.set_data( pd.DataFrame(np.zeros((len(catchments),5)), 
                                       columns=['code','reslist','lvs','cvi','cvs'], dtype="str"), index_col='code')
    ernst_drr.index = catchments.code
    for num, cat in enumerate(catchments.itertuples()):            
        ernst_drr.at[cat.code, 'code'] = str(cat.code)
        ernst_drr.at[cat.code, 'reslist'] = ' '.join([str(res) for res in resistance])
        ernst_drr.at[cat.code, 'lvs'] = ' '.join([str(depth) for depth in depths])
        ernst_drr.at[cat.code, 'cvi'] = str(infiltration_resistance)
        ernst_drr.at[cat.code, 'cvs'] = str(runoff_resistance)
    return ernst_drr     

def generate_paved( catchments=None, 
                    overflows=None,
                    sewer_areas=None,                                   
                    landuse=None, 
                    surface_level=None,
                    street_storage=None,
                    sewer_storage=None,
                    pump_capacity=None, 
                    meteo_areas=None,
                    zonalstats_alltouched=None):
    """ 
        Combine all data to form a complete PAVED definition. ALso the coordinates for the networ topology are included.
    Zonal statistics are applied to land use, to get the paved area. The classificition described in the notebook is assumed. From the elevation grid, the median value per catchment is assumed. Other parameters can be prescribed as as float (spatially uniform) or as a raster name, in which case the mean value per catchment is used.
    """
    
    all_touched=False if zonalstats_alltouched is None else zonalstats_alltouched        
        
    lu_rast, lu_affine = read_raster(landuse, static=True)
    lu_counts = zonal_stats(catchments, lu_rast, affine=lu_affine, categorical=True, all_touched=all_touched)   
    sl_rast, sl_affine = read_raster(surface_level, static=True)
    mean_elev = zonal_stats(catchments, sl_rast, affine=sl_affine, stats="median",all_touched=all_touched)         
    
    if isinstance( street_storage, str):                
        strs_rast, strs_affine = read_raster(street_storage, static=True)
        str_stors = zonal_stats(catchments, strs_rast, affine=strs_affine, stats="mean", all_touched=True)        
    elif isinstance( street_storage, int):
        street_storage = float(street_storage)
    if isinstance( sewer_storage, str):                
        sews_rast, sews_affine = read_raster(sewer_storage, static=True)
        sew_stors = zonal_stats(catchments, sews_rast, affine=sews_affine, stats="mean", all_touched=True)        
    elif isinstance( sewer_storage, int):
        sewer_storage = float(sewer_storage)    
    if isinstance(pump_capacity, str):     
        pump_rast, pump_affine = read_raster(pump_capacity, static=True)               
        pump_caps = zonal_stats(catchments, pump_rast, affine=pump_affine, stats="mean", all_touched=True)    
    elif isinstance( pump_capacity, int):
        pump_capacity = float(pump_capacity)    
        
    def update_dict(dict1, dict2):
        for i in dict2.keys():
            if i in dict1:
                dict1[i]+=dict2[i]
            else:
                dict1[i] = dict2[i]
        return dict1
    
    # get raster cellsize    
    px_area = lu_affine[0] * -lu_affine[4]
    paved_drr = ExtendedDataFrame(required_columns=['code'])
    if sewer_areas is not None:        
        # if the parameters area rasters, do the zonal statistics per sewage area as well.
        if isinstance( street_storage, str):                
            str_stors_sa = zonal_stats(sewer_areas, strs_rast,affine=strs_affine,stats="mean", all_touched=True)
        if isinstance( sewer_storage, str):                
            sew_stors_sa = zonal_stats(sewer_areas, sews_rast,affine=sews_affine,stats="mean", all_touched=True)
        if isinstance(pump_capacity, str):     
            pump_caps_sa = zonal_stats(sewer_areas, pump_rast,affine=pump_affine,stats="mean", all_touched=True)
        mean_sa_elev = zonal_stats(sewer_areas, sl_rast, affine=sl_affine, stats="median",all_touched=True)         
        
        # initialize the array of paved nodes, which should contain a node for all catchments and all overflows
        paved_drr.set_data( pd.DataFrame(np.zeros((len(catchments)+len(overflows),10)), 
columns=['code','area','mvlevel', 'streetstor', 'sewstor', 'pumpcap','meteostat','px', 'py', 'boundary'], dtype="str"), index_col='code')
        paved_drr.index = catchments.code.append(overflows.code)
        
        # find the paved area in the sewer areas
        for isew, sew in enumerate(sewer_areas.itertuples()):            
            pav_area = 0
            for cat_ind, cat in enumerate(catchments.itertuples()):                
                # if no rasterdata could be obtained for this catchment, skip it.
                if mean_elev[cat_ind]['median'] is None:
                   logger.warning('No rasterdata available for catchment %s' % cat.code)     
                   continue
                if(cat.geometry.intersects(sew.geometry)):
                    test_intersect = cat.geometry.intersection(sew.geometry)
                    #print(cat.Index+' '+sew.Index+' '+test_intersect.type)
                    if test_intersect.type =='LineString':
                        logger.warning('Intersection in %s contains of LineStrings, not polygons. Skipping. '% cat.code)
                        continue
                    if test_intersect.type=='GeometryCollection':                                                
                        numpol = 0
                        logger.info('Intersection in %s contains a GeometryCollection - splitting into polygons.'% cat.code)
                        for int_ft in test_intersect:                            
                            if int_ft.type == 'Polygon':
                                if numpol==0:                                                                
                                    intersecting_pixels = zonal_stats(int_ft, lu_rast, affine=lu_affine, categorical=True, all_touched=all_touched)[0]
                                else:                                    
                                    temp_int = zonal_stats(int_ft, lu_rast, affine=lu_affine, categorical=True, all_touched=all_touched)[0]
                                    intersecting_pixels = update_dict(intersecting_pixels, temp_int)
                                numpol += 1                        
                    else:
                        # find the paved area within the intersection and add it to the sewer area sum
                        intersecting_pixels = zonal_stats(cat.geometry.intersection(sew.geometry), lu_rast, affine=lu_affine, categorical=True, all_touched=all_touched)[0]
                    if intersecting_pixels=={}:
                        continue
                    if 14.0 not in intersecting_pixels:
                        logger.warning('%s/%s: no paved area in sewer area intersection!' % (sew.code, cat.code))
                        continue
                        
                    pav_pixels = intersecting_pixels[14.0]
                    pav_area += pav_pixels*px_area
                    # subtract it fromthe total paved area in this catchment, make sure at least 0 remains
                    lu_counts[cat_ind][14.0] -=  pav_pixels
                    if lu_counts[cat_ind][14.0] < 0: lu_counts[cat_ind][14.0]  = 0
            
            elev = mean_sa_elev[isew]['median']
            # find overflows related to this sewer area
            ovf = overflows[overflows.codegerelateerdobject==sew.code]           
            for ov in ovf.itertuples():                
                # find corresponding meteo-station
                tm = [m for m in meteo_areas.itertuples() if m.geometry.contains(sew.geometry.centroid)]
                ms = meteo_areas.iloc[0,:][0] if tm==[] else tm[0].code                         
                #ms = ms[0] if ms != [] else meteo_areas.iloc[0,:][0]                            
                # add prefix to the overflow id to create the paved-node id
                paved_drr.at[ov.code, 'code'] = str(ov.code)
                paved_drr.at[ov.code, 'area'] = str(pav_area * ov.fractie)
                paved_drr.at[ov.code, 'mvlevel'] = f'{elev:.2f}'         
                # if a float is given, a standard value is passed. If a string is given, a rastername is assumed to zonal statistics are applied.       
                if isinstance(street_storage, float):    
                    paved_drr.at[ov.code, 'streetstor'] = f'{street_storage:.2f}'
                else:
                    paved_drr.at[ov.code, 'streetstor'] = f'{str_stors_sa[isew]["mean"]:.2f}'                   
                if isinstance(sewer_storage, float):    
                    paved_drr.at[ov.code, 'sewstor'] = f'{sewer_storage:.2f}'
                else:
                    paved_drr.at[ov.code, 'sewstor'] = f'{sew_stors_sa[isew]["mean"]:.2f}'            
                if isinstance(pump_capacity, float):    
                    paved_drr.at[ov.code, 'pumpcap'] = f'{pump_capacity}'
                else:
                    paved_drr.at[ov.code, 'pumpcap'] = f'{pump_caps_sa[isew]["mean"]:.2f}'    
                paved_drr.at[ov.code,'meteostat'] = ms
                paved_drr.at[ov.code, 'px'] = f'{ov.geometry.coords[0][0]+10:.0f}'
                paved_drr.at[ov.code, 'py'] = f'{ov.geometry.coords[0][1]:.0f}'
                paved_drr.at[ov.code, 'boundary'] = ov.code              
    else:    
        # in this case only the catchments are taken into account. A node is created for every catchment nonetheless, but only nodes with a remaining area >0 are written.
        paved_drr.set_data( pd.DataFrame(np.zeros((len(catchments),10)), 
                                       columns=['code','area','mvlevel', 'streetstor', 'sewstor', 'pumpcap','meteostat', 'px', 'py', 'boundary'], dtype="str"), index_col='code')
        paved_drr.index = catchments.code
    
    for num, cat in enumerate(catchments.itertuples()):                    
        # if no rasterdata could be obtained for this catchment, skip it.
        if mean_elev[num]['median'] is None:
            logger.warning('No rasterdata available for catchment %s' % cat.code)     
            continue
        
        # find corresponding meteo-station
        tm = [m for m in meteo_areas.itertuples() if m.geometry.contains(cat.geometry.centroid)]
        ms = meteo_areas.iloc[0,:][0] if tm==[] else tm[0].code        
            
        elev = mean_elev[num]['median']
        paved_drr.at[cat.code, 'code'] = str(cat.code)
        paved_drr.at[cat.code, 'area'] = str(lu_counts[num][14]*px_area) if 14 in lu_counts[num] else '0'        
        paved_drr.at[cat.code, 'mvlevel'] = f'{elev:.2f}'         
        # if a float is given, a standard value is passed. If a string is given, a rastername is assumed to zonal statistics are applied.       
        if isinstance(street_storage, float):    
            paved_drr.at[cat.code, 'streetstor'] = f'{street_storage:.2f}'
        else:
            paved_drr.at[cat.code, 'streetstor'] = f'{str_stors[num]["mean"]:.2f}'                   
        if isinstance(sewer_storage, float):    
            paved_drr.at[cat.code, 'sewstor'] = f'{sewer_storage:.2f}'
        else:
            paved_drr.at[cat.code, 'sewstor'] = f'{sew_stors[num]["mean"]:.2f}'            
        if isinstance(pump_capacity, float):    
            paved_drr.at[cat.code, 'pumpcap'] = f'{pump_capacity}'
        else:
            paved_drr.at[cat.code, 'pumpcap'] = f'{pump_caps[num]["mean"]:.2f}'    
        paved_drr.at[cat.code,'meteostat'] = ms
        paved_drr.at[cat.code, 'px'] = f'{cat.geometry.centroid.coords[0][0]+10:.0f}'
        paved_drr.at[cat.code, 'py'] = f'{cat.geometry.centroid.coords[0][1]:.0f}'
        paved_drr.at[cat.code, 'boundary'] = cat.lateraleknoopcode                        
    return paved_drr   
       
def generate_greenhouse(catchments, landuse, surface_level, roof_storage, meteo_areas, zonalstats_alltouched=None):    
    """ 
        Combine all data to form a complete GREENHSE defintion. ALso the coordinates for the network topology are included.
    Zonal statistics are applied to land use, to get the paved area. The classificition described in the notebook is assumed. From the elevation grid, the median value per catchment is assumed. Other parameters can be prescribed as as float (spatially uniform) or as a raster name, in which case the mean value per catchment is used.
    """
    all_touched=False if zonalstats_alltouched is None else zonalstats_alltouched
        
    lu_rast, lu_affine = read_raster(landuse, static=True)
    lu_counts = zonal_stats(catchments, lu_rast, affine=lu_affine, categorical=all_touched)   
    rast, affine = read_raster(surface_level, static=True)
    mean_elev = zonal_stats(catchments, rast, affine=affine, stats="median", all_touched=all_touched) 
    # optional rasters
    if isinstance(roof_storage, str):                
        rast, affine = read_raster(roof_storage, static=True)
        roofstors = zonal_stats(catchments, rast, affine=affine, stats="mean", all_touched=True)    
    elif isinstance(roof_storage, int):
        roof_storage = float(roof_storage)
        
    # get raster cellsize    
    px_area = lu_affine[0] * -lu_affine[4]
    
    gh_drr = ExtendedDataFrame(required_columns=['code'])
    gh_drr.set_data( pd.DataFrame(np.zeros((len(catchments),8)), 
                                       columns=['code','area','mvlevel', 'roofstor', 'meteostat','px', 'py', 'boundary'], dtype="str"), index_col='code')
    gh_drr.index = catchments.code
    for num, cat in enumerate(catchments.itertuples()):    
        # if no rasterdata could be obtained for this catchment, skip it.
        if mean_elev[num]['median'] is None:
            logger.warning('No rasterdata available for catchment %s' % cat.code)     
            continue
                
        # find corresponding meteo-station
        tm = [m for m in meteo_areas.itertuples() if m.geometry.contains(cat.geometry.centroid)]
        ms = meteo_areas.iloc[0,:][0] if tm==[] else tm[0].code   
            
        elev = mean_elev[num]['median']
        gh_drr.at[cat.code, 'code'] = str(cat.code)
        gh_drr.at[cat.code, 'area'] = str(lu_counts[num][15]*px_area) if 15 in lu_counts[num] else '0'        
        gh_drr.at[cat.code, 'mvlevel'] = f'{elev:.2f}' 
        if isinstance(roof_storage, float):    
            gh_drr.at[cat.code, 'roofstor'] = f'{roof_storage:.2f}'
        else:
            gh_drr.at[cat.code, 'roofstor'] = f'{roofstors[num]["mean"]:.2f}'    
        gh_drr.at[cat.code, 'meteostat'] = ms
        gh_drr.at[cat.code, 'px'] = f'{cat.geometry.centroid.coords[0][0]+20:.0f}'
        gh_drr.at[cat.code, 'py'] = f'{cat.geometry.centroid.coords[0][1]:.0f}'
        gh_drr.at[cat.code, 'boundary'] = cat.lateraleknoopcode                        
    return gh_drr   

def generate_openwater(catchments, landuse, meteo_areas, zonalstats_alltouched=None):    
    """ 
        Combine all data to form a complete OPENWATE definotion. ALso the coordinates for the network topology are included.
    Zonal statistics are applied to land use, to get the open water area. The classificition described in the notebook is assumed. 
    """
    all_touched=False if zonalstats_alltouched is None else zonalstats_alltouched
    
    lu_rast, lu_affine = read_raster(landuse, static=True)
    lu_counts = zonal_stats(catchments, lu_rast, affine=lu_affine, categorical=True, all_touched=all_touched)   

    # get raster cellsize    
    px_area = lu_affine[0] * -lu_affine[4]
    
    ow_drr = ExtendedDataFrame(required_columns=['code'])
    ow_drr.set_data( pd.DataFrame(np.zeros((len(catchments),6)), 
                                       columns=['code','area','meteostat','px', 'py', 'boundary'], dtype="str"), index_col='code')
    ow_drr.index = catchments.code
    for num, cat in enumerate(catchments.itertuples()):    
        # find corresponding meteo-station
        tm = [m for m in meteo_areas.itertuples() if m.geometry.contains(cat.geometry.centroid)]
        ms = meteo_areas.iloc[0,:][0] if tm==[] else tm[0].code   
        
        ow_drr.at[cat.code, 'code'] = str(cat.code)
        ow_drr.at[cat.code, 'area'] = str(lu_counts[num][13]*px_area) if 13 in lu_counts[num] else '0'        
        ow_drr.at[cat.code, 'meteostat'] = ms
        ow_drr.at[cat.code, 'px'] = f'{cat.geometry.centroid.coords[0][0]-20:.0f}'
        ow_drr.at[cat.code, 'py'] = f'{cat.geometry.centroid.coords[0][1]:.0f}'
        ow_drr.at[cat.code, 'boundary'] = cat.lateraleknoopcode                        
    return ow_drr   

def generate_boundary(boundary_nodes, catchments, overflows=None):
    """
    Method to create boundary  nodes for RR.

    """
    if overflows is not None:
        numlats = len(catchments)+len(overflows)
    else:
        numlats = len(catchments)
    bnd_drr = ExtendedDataFrame(required_columns=['code'])
    bnd_drr.set_data( pd.DataFrame(np.zeros((numlats,3)), 
                                       columns=['code', 'px', 'py'], dtype="str"), index_col='code')                                        
    if overflows is not None:
        bnd_drr.index = catchments.code.append(overflows.code)
    else:
        bnd_drr.index = catchments.code
    for num, cat in enumerate(catchments.itertuples()):    
        # print(num, cat.code)
        if boundary_nodes[boundary_nodes['code']==cat.lateraleknoopcode].empty:
            #raise IndexError(f'{cat.code} not connected to a boundary node. Skipping.')
            logger.warning('%s not connected to a boundary node. Skipping.' % cat.code)
            continue
        bnd_drr.at[cat.code, 'code'] = cat.lateraleknoopcode        
        bnd_drr.at[cat.code, 'px']  = str(boundary_nodes[boundary_nodes['code']==cat.lateraleknoopcode]['geometry'].x.iloc[0]).strip()                                                                 
        bnd_drr.at[cat.code, 'py']  = str(boundary_nodes[boundary_nodes['code']==cat.lateraleknoopcode]['geometry'].y.iloc[0]).strip()
    if overflows is not None:
        logger.info('Adding overflows to the boundary nodes.')
        for num, ovf in enumerate(overflows.itertuples()):
            bnd_drr.at[ovf.code, 'code'] = ovf.code
            bnd_drr.at[ovf.code, 'px'] = str(ovf.geometry.coords[0][0])
            bnd_drr.at[ovf.code, 'py'] = str(ovf.geometry.coords[0][1])       
    return bnd_drr    

def generate_seepage(catchments, seepage_folder):    
    """
    Method to obtain catchment-average seepage fluxes from rasters. The time step is deduced from the raster filenames. 
    
    We assume seepage is read from Metaswap (m3 per cell). It needs to be converted to mm/day.
    """
    warnings.filterwarnings('ignore')
    file_list = os.listdir(seepage_folder)
    times = []    
    arr = np.zeros((len(file_list), len(catchments.code)))
    for ifile, file in tqdm(enumerate(file_list),total=len(file_list),desc='Reading seepage files'):
        array, affine, time = read_raster(os.path.join(seepage_folder, file))
        times.append(time)           
        stats = zonal_stats(catchments, array, affine=affine, stats="mean", all_touched=True)
        arr[ifile,:] = [s['mean'] for s in stats]        
    result = pd.DataFrame(arr,columns='sep_'+catchments.code)
    result.index = times 
    # convert units
    result_mmd = (result / (1e-3*(affine[0]*-affine[4])))/((times[2]-times[1]).total_seconds()/86400.)
    return result_mmd

def generate_precip(areas, precip_folder):
    """
    Method to obtain catchment-average seepage fluxes from rasters. The time step is deduced from the raster filenames.
    
    """
    warnings.filterwarnings('ignore')
    file_list = os.listdir(precip_folder)
    times = []        
    arr = np.zeros((len(file_list), len(areas.code)))
    for ifile, file in tqdm(enumerate(file_list),total=len(file_list),desc='Reading precipitation files'):
        array, affine, time = read_raster(os.path.join(precip_folder, file))
        times.append(time)                   
        stats = zonal_stats(areas, array, affine=affine, stats="mean", all_touched=True)
        arr[ifile,:]= [s['mean'] for s in stats]        
    result = pd.DataFrame(arr, columns='ms_'+areas.code)
    result.index = times 
    return result

def generate_evap(areas, evap_folder):    
    """
    Method to obtain catchment-average evaporation fluxes from rasters. The time step is deduced from the raster filenames. Since only one timeeries is allowed, the meteo areas are dissolved to a user specifield field.
    
    """
    warnings.filterwarnings('ignore')
    file_list = os.listdir(evap_folder)
    # aggregated evap
    areas['dissolve'] = 1
    agg_areas = areas.iloc[0:len(areas),:].dissolve(by='dissolve',aggfunc='mean')
    times = []    
    arr = np.zeros((len(file_list), 1))
    for ifile, file in tqdm(enumerate(file_list),total=len(file_list),desc='Reading evaporation files'):
        array, affine, time = read_raster(os.path.join(evap_folder, file))
        times.append(time)                  
        stats = zonal_stats(agg_areas, array, affine=affine, stats="mean",all_touched=True)
        arr[ifile,:] = [s['mean'] for s in stats]       
    result = pd.DataFrame(arr,columns=['ms_'+areas.iloc[0,0]])
    result.index = times 
    return result

def read_raster(file, static=False):
    """
    Method to read a raster. All rasterio types are accepted, plus IDF: in that case the iMod-package is used to read the IDF raster (IDF is cusomary for MODFLOW/SIMGRO models.)

    Parameters
    ----------
    file : raster
        
    static : BOOL, optional
        If static than no time information needs to be deduced.

    Returns
    -------
    rasterio grid and an affine object.        

    """
    if not static:         
        time = pd.Timestamp(os.path.split(file)[1].split('_')[1].split('.')[0])
    if file.lower().endswith('idf'):        
        dataset = imod.idf.open(file)
        header = imod.idf.header(file,pattern=None)        
        grid = dataset[0,0,:,:].values        
        affine =from_origin(header['xmin'], header['ymax'], header['dx'], header['dx'])                            
    else:        
        dataset = rasterio.open(file)
        affine = dataset.transform
        grid = dataset.read(1)
    if static:
        return grid, affine
    else:
        return grid, affine, time
        



from ctypes import POINTER, Structure, c_char, c_double, c_int
import os

import numpy as np
from shapely.geometry import LineString


# Definieer load structure
class meshgeomdim(Structure):

    _fields_ = [
        ('meshname', POINTER(c_char)),
        ('dim', c_int),
        ('numnode', c_int),
        ('numedge', c_int),
        ('numface', c_int),
        ('maxnumfacenodes', c_int),
        ('numlayer', c_int),
        ('layertype', c_int),
        ('nnodes', c_int),
        ('nbranches', c_int),
        ('ngeometry', c_int),
        ('epgs', c_int)
    ]

    def __repr__(self):
        strings = []
        for field, _ in self._fields_:
            value = getattr(self, field)
            strings.append(f'{field:15s}: {value}')

        return '\n'.join(strings)

class meshgeom(Structure):
    _fields_ = [
        ('edge_nodes', POINTER(c_int)),
        ('face_nodes', POINTER(c_int)),
        ('edge_faces', POINTER(c_int)),
        ('face_edges', POINTER(c_int)),
        ('face_links', POINTER(c_int)),
        ('nnodex', POINTER(c_double)),
        ('nnodey', POINTER(c_double)),
        ('nedge_nodes', POINTER(c_int)),
        ('nbranchlengths', POINTER(c_double)),
        ('nbranchgeometrynodes', POINTER(c_int)),
        ('ngeopointx', POINTER(c_double)),
        ('ngeopointy', POINTER(c_double)),
        ('nbranchorder', POINTER(c_double)),
        ('branchidx', POINTER(c_int)),
        ('branchoffsets', POINTER(c_double)),
        #('edge_branchidx', POINTER(c_int)),
        #('edge_branchoffsets', POINTER(c_double)),        
        ('nodex', POINTER(c_double)),
        ('nodey', POINTER(c_double)),
        ('nodez', POINTER(c_double)),
        ('edgex', POINTER(c_double)),
        ('edgey', POINTER(c_double)),
        ('edgez', POINTER(c_double)),
        ('facex', POINTER(c_double)),
        ('facey', POINTER(c_double)),
        ('facez', POINTER(c_double)),
        ('layer_zs', POINTER(c_int)),
        ('interface_zs', POINTER(c_int))
    ]

    def __init__(self, geometries):
        """
        Constructor
        """
        
        self.meshgeomdim = geometries

        # This dictionary contains some extra variables for 1d meshes
        self.description1d = {
            'mesh1d_node_ids': [],
            'mesh1d_node_long_names': [],
            'network_node_ids': [],
            'network_node_long_names': [],
            'network_branch_ids': [],
            'network_branch_long_names': []
        }

        self._meta_ = {
            'nodex': {'ctype': c_double, 'size': ['numnode'], 'allocated': False},
            'nodey': {'ctype': c_double, 'size': ['numnode'], 'allocated': False},
            'nodez': {'ctype': c_double, 'size': ['numnode'], 'allocated': False},
            'branchoffsets': {'ctype': c_double, 'size': ['numnode'], 'allocated': False},
            'branchidx': {'ctype': c_int, 'size': ['numnode'], 'allocated': False},            
            #'edge_branchoffsets': {'ctype': c_double, 'size': ['numedge'], 'allocated': False},
            #'edge_branchidx': {'ctype': c_int, 'size': ['numedge'], 'allocated': False},
            'nnodex': {'ctype': c_double, 'size': ['nnodes'], 'allocated': False},
            'nnodey': {'ctype': c_double, 'size': ['nnodes'], 'allocated': False},
            'ngeopointx': {'ctype': c_double, 'size': ['ngeometry'], 'allocated': False},
            'ngeopointy': {'ctype': c_double, 'size': ['ngeometry'], 'allocated': False},
            'nbranchlengths': {'ctype': c_double, 'size': ['nbranches'], 'allocated': False},
            'nbranchorder': {'ctype': c_double, 'size': ['nbranches'], 'allocated': False},
            'edgex': {'ctype': c_double, 'size': ['numedge'], 'allocated': False},
            'edgey': {'ctype': c_double, 'size': ['numedge'], 'allocated': False},
            'edgez': {'ctype': c_double, 'size': ['numedge'], 'allocated': False},
            'nedge_nodes': {'ctype': c_int, 'size': ['nbranches', 2], 'allocated': False},
            'nbranchgeometrynodes': {'ctype': c_int, 'size': ['nbranches'], 'allocated': False},
            'facex': {'ctype': c_double, 'size': ['numface'], 'allocated': False},
            'facey': {'ctype': c_double, 'size': ['numface'], 'allocated': False},
            'facez': {'ctype': c_double, 'size': ['numface'], 'allocated': False},
            'face_nodes': {'ctype': c_int, 'size': ['numface', 'maxnumfacenodes'], 'allocated': False},
            'edge_nodes': {'ctype': c_int, 'size': ['numedge', 2], 'allocated': False},
            'edge_faces': {'ctype': c_int, 'size': ['numedge', 2], 'allocated': False},
            'face_edges': {'ctype': c_int, 'size': ['numface', 'maxnumfacenodes'], 'allocated': False}
        }

    def get_dimensions(self, var):
        return tuple(getattr(self.meshgeomdim, fac) if isinstance(fac, str) else fac for fac in self._meta_[var]['size'])

    def get_size(self, var):
        return np.prod(self.get_dimensions(var))

    def allocate(self, var):
        ctype = self._meta_[var]['ctype']
        size = self.get_size(var)
        setattr(self, var, (ctype * size)())
        self._meta_[var]['allocated'] = True

    def is_allocated(self, var):
        return self._meta_[var]['allocated']

    def get_values(self, var, as_array=False, size=None):
        if not self.is_allocated(var):
            return None
        size = self.get_size(var) if size is None else size
        values = [getattr(self, var)[i] for i in range(size)]
        if as_array:
            values = np.reshape(values, self.get_dimensions(var))
        return values

    def set_values(self, var, values):
        # First allocate
        self.allocate(var)

        size = self.get_size(var)
        if size != len(values):
            raise ValueError(f'Size of values ({len(values)}) does not match allocated size ({size}) for "{var}".')
        for i, value in enumerate(values):
            getattr(self, var)[i] = value


    def add_values(self, var, values):
        
        # If not yet allocated, don't add but set
        if not self.is_allocated(var):
            self.set_values(var, values)
        
        # Get old values
        old_values = self.get_values(var, size=(self.get_size(var)-len(values)))
        
        # First allocate
        self.allocate(var)
        size = self.get_size(var)
        
        if size != (len(values) + len(old_values)):
            raise ValueError(f'Size of values ({len(values) + len(old_values)}) does not match allocated size ({size})')

        for i, value in enumerate(old_values + values):
            getattr(self, var)[i] = value

    def add_from_other(self, geometries):
        """
        Method to merge mesh with another mesh
        """

        # If the current mesh is empty, copy the maxfacenumnodes
        if self.empty():
            self.meshgeomdim.maxnumfacenodes = geometries.meshgeomdim.maxnumfacenodes
        # If the maxnumfacenodes is not equal, raise an error. Merging two diffently shaped meshes not implemented
        if self.meshgeomdim.maxnumfacenodes != geometries.meshgeomdim.maxnumfacenodes:
            raise NotImplementedError('The maximum number of face nodes differs between the meshes.')

        # Get the counter offset for the nodes
        startnode = self.meshgeomdim.numnode
        startbranch = self.meshgeomdim.nbranches
        
        # Add dimensions. Sum the new dimensions with the old ones
        for dimension in ['numnode', 'numedge', 'numface', 'nnodes', 'ngeometry', 'nbranches']:
            new = getattr(self.meshgeomdim, dimension) + getattr(geometries.meshgeomdim, dimension)
            setattr(self.meshgeomdim, dimension, new)
        
        # Add variables. Add all new data
        for var in ['nodex', 'nodey', 'nodez', 'facex', 'facey', 'facez', 'nnodex', 'nnodey',
                    'nbranchlengths', 'nbranchorder', 'ngeopointx', 'ngeopointy', 'nbranchgeometrynodes']:
            if geometries.is_allocated(var):
                self.add_values(var, geometries.get_values(var))

        # For variables with indexes. For the indexes, add the old start node
        for indexvar in ['edge_nodes', 'face_nodes', 'branchidx', 'branchoffsets']:
            if geometries.is_allocated(indexvar):
                self.add_values(indexvar, [i + startnode for i in geometries.get_values(indexvar)]) 

        # Network indexvar
        indexvar = 'nedge_nodes'
        if geometries.is_allocated(indexvar):
            self.add_values(indexvar, [i + startbranch for i in geometries.get_values(indexvar)]) 

        # Network descriptive variables
        for var in ['network_node_ids', 'network_node_long_names', 'network_branch_ids',
                    'network_branch_long_names', 'mesh1d_node_ids', 'mesh1d_node_long_names']:
            self.description1d[var] += geometries.description1d[var]
        
        
    def process_1d_network(self):
        """
        Determine x, y locations of 1d network
        """
        assert self.meshgeomdim.dim == 1

        # Read network geometry nodes. Create an inverted list to pop
        ngeom = list(zip(self.get_values('ngeopointx'), self.get_values('ngeopointy')))

        # Get branch data
        nbranchnames = self.description1d['network_branch_ids']
        ngeometrynodes = self.get_values('nbranchgeometrynodes')
        branchids = self.get_values('branchidx', as_array=True)
        offsets = self.get_values('branchoffsets', as_array=True)
        
        # Collect branches
        branches = {}
        schematised = {}
        crds = []
        for i, (name, nnodes) in enumerate(zip(nbranchnames, ngeometrynodes)):
            # Create linestring for network branch
            linestring = LineString([ngeom.pop(0) for _ in range(nnodes)])
            branches[name.strip()] = linestring
            # Determine mesh node position on network branch
            branchoffsets = offsets[branchids == (i + 1)]
            meshcrds = [linestring.interpolate(offset).coords[0] for offset in branchoffsets]
            crds.extend(meshcrds[:])
            # Determine if a start or end coordinate needs to be added for constructing a complete LineString
            if not np.isclose(branchoffsets[0], 0.0):
                meshcrds = [linestring.coords[0]] + meshcrds
            if not np.isclose(branchoffsets[-1], linestring.length):
                meshcrds = meshcrds + [linestring.coords[-1]]
            schematised[name.strip()] = LineString(meshcrds)

        # Add values to mesh
        nodex, nodey = list(zip(*crds))
        self.set_values('nodex', nodex)
        self.set_values('nodey', nodey)

        return schematised, branches

    def empty(self):
        """Determine whether mesh is empty, based on number of nodes"""
        return not bool(self.meshgeomdim.numnode)

    def get_nodes(self):
        """
        Return nodes.
        
        Returns
        -------
        np.ndarray
            Numpy array with x and y coordinates of nodes.
        """

        # Get nodes
        nodes = np.c_[
            np.array(self.get_values('nodex')),
            np.array(self.get_values('nodey'))
        ]

        return nodes

    def get_nodes_for_branch(self, branchid):
        """
        Return index of nodes on branch(es).

        Note that each nodes belongs to a single branch in the network.
        This can provide unexpected results on intersections.
        
        Parameters
        ----------
        branchid : str or list
            Branchid for which to return the nodes.
        
        Returns
        -------
        np.ndarray
            boolean array with True for nodes that are on the branch(es)
        """        
        if branchid is None:
            return np.ones(len(self.get_values('nodex')), dtype=bool)

        # Convert to list if needed
        if isinstance(branchid, str):
            branchid = [branchid]
        # Get the ids (integers) of the branch names given by the user
        branchidx = np.where(np.isin(self.description1d['network_branch_ids'], branchid))[0] + 1
        # Select which of the nodes are in the branches
        idx = np.isin(self.get_values('branchidx'), branchidx)
        
        return idx

    def get_segments(self):

        # Read nodes and links from src
        nodes = self.get_nodes()
        # Get links
        links = self.get_values('edge_nodes', as_array=True)

        return nodes[links - 1]


    def get_faces(self, geometry='exterior'):
        """
        Get cells from 2d mesh

        There are three options for returning coordinates:
        - "center": can be the circumcenter (default) or the centroid if the face centers are moved by the user
        - "centroid": the centroid, which is calculated (again) in this function
        - "exterior" (default): the coordinates of the face edges
        """

        assert self.meshgeomdim.dim == 2
        
        # Read nodes and links from src
        nodes = self.get_nodes()

        if geometry == 'center':
            # Can be both the centroid and the circumcenter
            return np.c_[self.get_values('facex'), self.get_values('facey')]

        elif geometry in ['exterior', 'centroid']:

            # Get face nodes
            face_nodes = self.get_values('face_nodes', as_array=True)
            nanvalues = (face_nodes != -999).sum(axis=1).astype(int)
            unique = np.unique(nanvalues)
            
            # If all cells have the same number of nodes
            if len(unique) == 1:
                faces = nodes[face_nodes - 1]
                

            # Else, combine a list for all nodes
            else:
                faces = [None] * len(face_nodes)
                for n in unique:
                    ncells = nodes[face_nodes[(nanvalues == n), :n] - 1]
                    where = np.where(nanvalues == n)[0]
                    for i, cell in zip(where, ncells):
                        faces[i] = cell

            if geometry == 'exterior':
                return faces
            
            # Determine the centroids by averaging the exteriors
            if geometry == 'centroid':
                return np.vstack([f.mean(axis=0) for f in faces])
        
        else:
            raise ValueError(f'Geometry "{geometry}" not recognized. Pick "center", "centroid" or "exterior"')
            

        



import logging

import geopandas as gpd
import numpy as np
import pandas as pd
from scipy.spatial import KDTree

from delft3dfmpy.converters import hydamo_to_dflowfm
from delft3dfmpy.core import checks
from delft3dfmpy.datamodels.common import ExtendedGeoDataFrame
from shapely.geometry import LineString, MultiPolygon, Point, Polygon

logger = logging.getLogger(__name__)


class StructuresIO:

    def __init__(self, structures):
        self.structures = structures

    def pumps_from_hydamo(self, pompen, sturing, gemalen):
        """
        Method to generate dflowfm pumps from hydamo. Three dataframes are needed:
        pompen (pumps), sturing (control) and gemalen (pump houses).
        """
        # Convert to dflowfm input
        geconverteerd = hydamo_to_dflowfm.generate_pumps(pompen, sturing, gemalen)
        # Add to dict
        for pump in geconverteerd.itertuples():
            self.structures.add_pump(
                id=pump.code,
                branchid=pump.branch_id,
                chainage=pump.branch_offset,
                orientation='positive',
                numstages=1,
                controlside='suctionSide',
                capacity=pump.maximalecapaciteit,
                startlevelsuctionside=pump.startlevelsuctionside,
                stoplevelsuctionside=pump.stoplevelsuctionside                
            )

    def orifices_from_hydamo(self, orifices):
        """
        Method to generate dflowfm orifices.
        """
        # Convert to dflowfm input
        geconverteerd = hydamo_to_dflowfm.generate_orifices(orifices)
        # Add to dict
        for orifice in geconverteerd.itertuples():
            self.structures.add_orifice(
                id=orifice.code,
                branchid=orifice.branch_id,
                chainage=orifice.branch_offset,                
                crestlevel=orifice.laagstedoorstroomhoogte,
                crestwidth=orifice.laagstedoorstroombreedte,
                gateloweredgelevel=orifice.schuifhoogte,
                uselimitflowpos=orifice.uselimitflow,
                limitflowpos=orifice.limitflow,
                uselimitflowneg=orifice.uselimitflow,
                limitflowneg=orifice.limitflow,
                corrcoeff=orifice.afvoercoefficient                               
            )
            
    def weirs_from_hydamo(self, weirs, yz_profiles=None, parametrised_profiles=None, afsluitmiddel=None, sturing=None):
        """
        Method to convert dflowfm weirs from hydamo weirs.
        
        Split the HyDAMO weirs object into a subset with 'regular weirs' and 'universal weirs'. 
       
        For now the assumption is that weirs for which a profile is defined, are universal, other weirs are regular.
               
        """
        # Convert to dflowfm input
        
        regular = ExtendedGeoDataFrame(geotype=Point, required_columns=weirs.required_columns)
        index = np.zeros((len(weirs.code)))
        if yz_profiles is not None:
            if 'codegerelateerdobject' in yz_profiles:            
                index[np.isin(weirs.code , np.asarray(yz_profiles.codegerelateerdobject))]=1
        if parametrised_profiles is not None:
            if 'codegerelateerdobject' in parametrised_profiles:            
                index[np.isin(weirs.code , np.asarray(parametrised_profiles.codegerelateerdobject))]=1
           
        regular.set_data(weirs[index==0], index_col='code',check_columns=True)        
        regular_geconverteerd = hydamo_to_dflowfm.generate_weirs(regular, afsluitmiddel=afsluitmiddel)        
        regular.set_data(weirs, index_col='code',check_columns=True)        
        
        # Add to dict
        for weir in regular_geconverteerd.itertuples():
            self.structures.add_weir(
                id=weir.code,
                branchid=weir.branch_id,
                chainage=weir.branch_offset,                
                crestlevel=weir.laagstedoorstroomhoogte,
                crestwidth=weir.laagstedoorstroombreedte,
                corrcoeff=weir.afvoercoefficient
            )
    
        universal = ExtendedGeoDataFrame(geotype=Point, required_columns=weirs.required_columns)
        universal.set_data(weirs[index==1], index_col='code',check_columns=True)                                      
        universal_geconverteerd = hydamo_to_dflowfm.generate_uweirs(universal,yz_profiles=yz_profiles, parametrised_profiles=parametrised_profiles)
        
        if universal_geconverteerd.empty:
            return("No profile detected for universal weir)")
        # Add to dict
        for uweir in universal_geconverteerd.itertuples():
            self.structures.add_uweir(
                id=uweir.code,
                branchid=uweir.branch_id,
                chainage=uweir.branch_offset,                
                crestlevel=uweir.laagstedoorstroomhoogte,                
                numlevels=uweir.numlevels,
                yvalues=uweir.yvalues,
                zvalues=uweir.zvalues,
                allowedflowdir='both',
                dischargecoeff=uweir.afvoercoefficient                
            )    
        
    def bridges_from_hydamo(self, bridges, yz_profiles=None, parametrised_profiles=None):
        """
        Method to convert dflowfm bridges from hydamo bridges.
        """
        # Convert to dflowfm input
        geconverteerd = hydamo_to_dflowfm.generate_bridges(bridges, yz_profiles=yz_profiles, parametrised_profiles=parametrised_profiles)
        # Add to dict
        for bridge in geconverteerd.itertuples():
            self.structures.add_bridge(
                id=bridge.code,
                branchid=bridge.branch_id,
                chainage=bridge.branch_offset,
                length=bridge.lengte,
                bedlevel=bridge.bedlevel,
                upperheight=bridge.hoogtebovenzijde,
                lowerheight=bridge.hoogteonderzijde,                
                crosssection=bridge.crosssection,
                inletlosscoeff=bridge.intreeverlies,
                outletlosscoeff=bridge.uittreeverlies,
                frictiontype=hydamo_to_dflowfm.roughness_gml[bridge.ruwheidstypecode],
                frictionvalue=bridge.ruwheidswaarde
            )
            

    def culverts_from_hydamo(self, culverts, afsluitmiddel=None):
        """
        Method to convert dflowfm weirs from hydamo weirs.
        """
        # Convert to dflowfm input
        geconverteerd = hydamo_to_dflowfm.generate_culverts(culverts, afsluitmiddel)

        # Add to dict
        for culvert in geconverteerd.itertuples():            
                self.structures.add_culvert(
                    id=culvert.code,
        	        branchid=culvert.branch_id,
        	        chainage=culvert.branch_offset,
        	        leftlevel=culvert.hoogtebinnenonderkantbovenstrooms,
        	        rightlevel=culvert.hoogtebinnenonderkantbenedenstrooms,
        	        crosssection=culvert.crosssection,
        	        length=culvert.lengte,#geometry.length,
        	        inletlosscoeff=culvert.intreeverlies,
        	        outletlosscoeff=culvert.uittreeverlies,
                    allowedflowdir=culvert.allowedflowdir,
                    valveonoff=culvert.valveonoff,
                    numlosscoeff=culvert.numlosscoeff, 
                    valveopeningheight=culvert.valveopeningheight,
                    relopening=culvert.relopening,
                    losscoeff=culvert.losscoeff,                                    
                    frictiontype=hydamo_to_dflowfm.roughness_gml[culvert.ruwheidstypecode],
                    frictionvalue=culvert.ruwheidswaarde
                )
        
    def compound_structures(self, idlist, structurelist):
        """
        Method to add compound structures to the model.
        
        """
        geconverteerd = hydamo_to_dflowfm.generate_compounds(idlist, structurelist, self.structures)
        
         # Add to dict
        for compound in geconverteerd.itertuples():
            self.structures.add_compound(
                id=compound.code,
                numstructures=compound.numstructures,
    	        structurelist=compound.structurelist    	        
            )
        
                    

class CrossSectionsIO:

    def __init__(self, crosssections):
        self.crosssections = crosssections

    def from_hydamo(self, dwarsprofielen, parametrised=None, branches=None):
        """
        Method to add cross section from hydamo files. Two files
        can be handed to the function, the cross section file (dwarsprofiel) and the
        parametrised file (normgeparametriseerd). The
        hierarchical order is 1. dwarsprofiel, 2. normgeparametriseerd.
        Each branch will be assigned a profile following this order. If parametrised
        and standard are not given, branches can be without cross section. In that case
        a standard profile should be assigned
        """

        # first, make a selection as to use only the dwarsprofielen/parametrised that are related to branches, not structures
        if dwarsprofielen is not None and not dwarsprofielen.empty:
            if 'codegerelateerdobject' not in dwarsprofielen:
                dwarsprofielen['codegerelateerdobject'] = np.empty((len(dwarsprofielen)))*np.nan
            dp_branches = ExtendedGeoDataFrame(geotype=LineString, columns = dwarsprofielen.required_columns+['codegerelateerdobject'])
            dp_branches.set_data(dwarsprofielen[dwarsprofielen.codegerelateerdobject.isna()], index_col='code', check_columns=True)
        else:     
            dp_branches = ExtendedGeoDataFrame(geotype=LineString)
            
        if parametrised is not None and not parametrised.empty:
            if 'codegerelateerdobject' not in parametrised:
                parametrised['codegerelateerdobject'] = np.empty((len(parametrised)))*np.nan
            if len(parametrised)>0:        
                par_branches = ExtendedGeoDataFrame(geotype=LineString, columns = parametrised.required_columns+['codegerelateerdobject'])
                par_branches.set_data(parametrised[parametrised.codegerelateerdobject.isna()], index_col='code', check_columns=True)
            else:
                par_branches = ExtendedGeoDataFrame(geotype=LineString, columns = parametrised.required_columns+['codegerelateerdobject'])
            # Assign cross-sections to branches
            nnocross = len(self.crosssections.get_branches_without_crosssection())
            logger.info(f'Before adding the number of branches without cross section is: {nnocross}.')
        else:
            par_branches = ExtendedGeoDataFrame(geotype=LineString)
            
        if not dp_branches.empty:
            # 1. Collect cross sections from 'dwarsprofielen'
            crosssections = hydamo_to_dflowfm.dwarsprofiel_to_yzprofiles(dp_branches, branches)               
        
            for name, css in crosssections.items():
                # Add definition
                self.crosssections.add_yz_definition(yz=css['yz'], thalweg=css['thalweg'], name=name, roughnesstype=css['ruwheidstypecode'], roughnessvalue=css['ruwheidswaarde'])
                # Add location
                self.crosssections.add_crosssection_location(branchid=css['branchid'], chainage=css['chainage'], definition=name)
        
        # Check the number of branches with cross sections
        no_crosssection = self.crosssections.get_branches_without_crosssection()
        nnocross = len(no_crosssection)
        logger.info(f'After adding \'dwarsprofielen\' the number of branches without cross section is: {nnocross}.')
        if (nnocross == 0):
            print('No further branches without a profile.')
        elif par_branches.empty:
            print('No parametrised crossections available for branches.')
        else: 
            # Derive norm cross sections for norm parametrised
            crosssections = hydamo_to_dflowfm.parametrised_to_profiles(par_branches, no_crosssection)
            # Get branch information
            branchdata = self.crosssections.dflowfmmodel.network.branches.loc[list(crosssections.keys())]
            branchdata['chainage'] = branchdata.length / 2.
        
            # Add cross sections
            for branchid, css in crosssections.items():
                chainage = branchdata.at[branchid, 'chainage']
                    
                if css['type'] == 'rectangle':
                    name = self.crosssections.add_rectangle_definition(
                        height=css['height'],
                        width=css['width'],
                        closed=css['closed'],
                        roughnesstype=css['ruwheidstypecode'],
                        roughnessvalue=css['ruwheidswaarde']
                    )
    
                if css['type'] == 'trapezium':
                    name = self.crosssections.add_trapezium_definition(
                        slope=css['slope'],
                        maximumflowwidth=css['maximumflowwidth'],
                        bottomwidth=css['bottomwidth'],
                        closed=css['closed'],
                        roughnesstype=css['ruwheidstypecode'],
                        roughnessvalue=css['ruwheidswaarde']
                    ) 
    
                # Add location
                self.crosssections.add_crosssection_location(branchid=branchid, chainage=chainage, definition=name, shift=css['bottomlevel'])

            nnocross = len(self.crosssections.get_branches_without_crosssection())
            logger.info(f'After adding \'normgeparametriseerd\' the number of branches without cross section is: {nnocross}.')
        
        # Assign cross-sections to structures        
        nnocross = len(self.crosssections.get_structures_without_crosssection())        
        logger.info(f'Before adding the number of structures without cross section is: {nnocross}.')    
        
        if dwarsprofielen is not None:
            # and subsets for structures    
            dp_structures = ExtendedGeoDataFrame(geotype=LineString)
            dp_structures.set_data(dwarsprofielen[~dwarsprofielen.codegerelateerdobject.isna()], index_col='code', check_columns=True)
        
            # 1. Collect cross sections from 'dwarsprofielen'
            crosssections = hydamo_to_dflowfm.dwarsprofiel_to_yzprofiles(dp_structures, None)               
        
            for name, css in crosssections.items():        
                self.crosssections.add_yz_definition(yz=css['yz'], thalweg=css['thalweg'], name=name, roughnesstype=css['ruwheidstypecode'], roughnessvalue=css['ruwheidswaarde'])
        
        if parametrised is not None:
            par_structures = ExtendedGeoDataFrame(geotype=LineString, columns = parametrised.required_columns + ['codegerelateerdobject'])
            par_structures.set_data(parametrised[~parametrised.codegerelateerdobject.isna()], index_col='code', check_columns=True)            
        else:
            par_structures = ExtendedGeoDataFrame(geotype=LineString)
      
        no_crosssection = self.crosssections.get_structures_without_crosssection()        
        nnocross = len(no_crosssection)
        logger.info(f'After adding \'dwarsprofielen\' the number of branches without cross section is: {nnocross}.')
        if (nnocross == 0):
            print('No further structures without a profile.')
        elif par_structures.empty:
            print('No parametrised crosssections available for structures.')
        else:                 
            # Derive norm cross sections for norm parametrised
            crosssections = hydamo_to_dflowfm.parametrised_to_profiles(par_structures,[])                
            # Add cross section definitions
            for strucid, css in crosssections.items():             
                        
                    if css['type'] == 'rectangle':
                        name = self.crosssections.add_rectangle_definition(
                            height=css['height'],
                            width=css['width'],
                            closed=css['closed'],
                            roughnesstype=css['ruwheidstypecode'],
                            roughnessvalue=css['ruwheidswaarde'],
                            name=strucid,                        
                        )
        
                    if css['type'] == 'trapezium':
                        name = self.crosssections.add_trapezium_definition(
                            slope=css['slope'],
                            maximumflowwidth=css['maximumflowwidth'],
                            bottomwidth=css['bottomwidth'],
                            closed=css['closed'],
                            roughnesstype=css['ruwheidstypecode'],
                            roughnessvalue=css['ruwheidswaarde'],
                            name=strucid
                        ) 
                        
        nnocross = len(self.crosssections.get_structures_without_crosssection())
        logger.info(f'After adding \'normgeparametriseerd\' the number of structures without cross section is: {nnocross}.')
            
class ExternalForcingsIO:

    def __init__(self, external_forcings):
        self.external_forcings = external_forcings

    def from_hydamo(self, boundary_conditions):

        # Read from Hydamo
        bcdct = hydamo_to_dflowfm.generate_boundary_conditions(boundary_conditions, self.external_forcings.dflowfmmodel.network.schematised)

        # Add all items
        for key, item in bcdct.items():
            self.external_forcings.add_boundary_condition(key, item['geometry'], item['bctype'], item['value'], branchid=item['branchid'])
            # # Add to dataframe         
            # self.external_forcings.boundaries.loc[key] = item
            # Check if a 1d2d link should be removed
            #self.external_forcings.dflowfmmodel.network.links1d2d.check_boundary_link(self.external_forcings.boundaries.loc[key])

    def read_laterals(self, locations, lateral_discharges=None, rr_boundaries=None):
        """
        Process laterals

        Parameters
        ----------
        locations: gpd.GeoDataFrame
            GeoDataFrame with at least 'geometry' (Point) and the column 'code'
        lateral_discharges: pd.DataFrame
            DataFrame with lateral discharges. The index should be a time object (datetime or similar).
        rr_boundaries: pd.DataFrame
            DataFrame with RR-catchments that are coupled 
        """

        if rr_boundaries is None: rr_boundaries = []
        # Check argument
        checks.check_argument(locations, 'locations', gpd.GeoDataFrame, columns=['geometry'])
        if lateral_discharges is not None:
            checks.check_argument(lateral_discharges, 'lateral_discharges', pd.DataFrame)

        # Check if network has been loaded
        network1d = self.external_forcings.dflowfmmodel.network.mesh1d
        if not network1d.meshgeomdim.numnode:
            raise ValueError('1d network has not been generated or loaded. Do this before adding laterals.')

        # in case of 3d points, remove the 3rd dimension
        locations['geometry2'] = [Point([point.geometry.x, point.geometry.y]) for _,point in locations.iterrows()]    
        locations.drop('geometry', inplace=True, axis=1)
        locations.rename(columns={'geometry2':'geometry'}, inplace=True)
        
        # Find nearest 1d node per location and find the nodeid
        #lateral_crds = np.vstack([loc.geometry.coords[0] for loc in locations.itertuples()])             
        #nodes1d = network1d.get_nodes()
        #get_nearest = KDTree(nodes1d)
        #_, nearest_idx = get_nearest.query(lateral_crds[:,0:2])
                
        # Get time series and add to dictionary
        #for nidx, lateral in zip(nearest_idx, locations.itertuples()):
        for lateral in locations.itertuples():
            # crd = nodes1d[nearest_idx]
            #nid = f'{nodes1d[nidx][0]:g}_{nodes1d[nidx][1]:g}'
            
            # Check if a time is provided for the lateral
            if lateral.code in rr_boundaries:                
                # Add to dictionary
                self.external_forcings.laterals[lateral.code] = {
                    'branchid': lateral.branch_id,
                    'branch_offset': str(lateral.branch_offset)                    
                }
            else:
                if lateral_discharges is None:
                    logger.warning(f'No lateral_discharges provied. {lateral.code} expects them. Skipping.')
                    continue
                else:
                    if lateral.code not in lateral_discharges.columns:
                        logger.warning(f'No data found for {lateral.code}. Skipping.')
                        continue
                    
                # Get timeseries
                series = lateral_discharges.loc[:, lateral.code]
                
                # Add to dictionary
                self.external_forcings.laterals[lateral.code] = { 
                    'branchid': lateral.branch_id,
                    'branch_offset': str(lateral.branch_offset), 
                    'timeseries': series            
                }                                   
        
        



# coding: latin-1
import os
import re
import shutil
import datetime
import warnings
import numpy as np
from shapely.geometry import box, LineString, Point
import pandas as pd
import csv
import delft3dfmpy
from delft3dfmpy.core.geometry import orthogonal_line

import logging

logger = logging.getLogger(__name__)

class DFlowRRWriter:
    """Writer for RR files"""

    def __init__(self, rrmodel, output_dir, name):
        self.rrmodel = rrmodel
        # self.geometries = parent.geometries
        # self.boundary_conditions = parent.boundary_conditions
        self.output_dir = os.path.join(output_dir, 'rr')
        self.name = name

        self.version = delft3dfmpy.DFlowFMWriter.version
        
        if os.path.exists(self.output_dir):
            shutil.rmtree(self.output_dir)

        self.d3b_parameters = rrmodel.d3b_parameters
        
        self.precip_df  = {}
        self.run_dimrpad = rrmodel.dimr_path
        
    def write_all(self):  # write all RR files
        """
        Wrapper method to write all components
        """        
        # copy all 'standard' files from the resources folder
        self.copyRRFiles()

        self.write_topology()
        
        self.write_unpaved()
        
        self.write_paved()
        
        self.write_greenhouse()
        
        self.write_openwater()
        
        self.write_meteo()
        
        self.write_coupling()

        # Change mdu parameters
        for parameter, value in self.d3b_parameters.items():
            self.change_d3b_parameter(parameter, value)
        
    def copyRRFiles(self):
        """
        Method to copy the RR-files from resources.

        Returns
        -------
        bool
            Succesfull or not.

        """
        srcRR = os.path.join(os.path.dirname(__file__), 'resources', 'RRmodel')
        targetRR = os.path.join(self.output_dir)
        shutil.copytree(srcRR, targetRR)
        return True

    def write_topology(self):
        """
        Wrapper to write the topolgy files for RR. The following files are written:
            3B_NOD.TP: IDs and coordinates of all nodes: unpaved, paved, greenhouse, open water and boundary. Only nodes with a total area of >0 are written. Nodes are placed up to 20 m to the east and west of the catchment centroid.
            3B_LINK.TP: Links between all nodes and the correct boundaries.
            BOUND3B.3B: coupling information for every RR boundary.
            BoundaryConditions.bc: see bound3b.3b
        Returns
        -------
        None.

        """
        filepath = os.path.join(self.output_dir, '3B_NOD.TP')
        with open(filepath, 'w') as f:

            # Unpaved nodes
            if any(self.rrmodel.unpaved.unp_nodes):
                for _, dct in self.rrmodel.unpaved.unp_nodes.items():
                    if np.sum([float(d) for d in dct['ar'].split(' ')]) > 0.0:                    
                        f.write('NODE id \''+dct['id']+'\' nm \''+dct['id']+'\' ri \'-1\' mt 1 \'2\' nt 44 ObID \'3B_UNPAVED\' px '+dct['px']+' py '+dct['py']+' node\n')
                              
            # Paved nodes
            if any(self.rrmodel.paved.pav_nodes):
                for _, dct in self.rrmodel.paved.pav_nodes.items():
                    if float(dct['ar']) > 0.0:                    
                        f.write('NODE id \''+dct['id']+'\' nm \''+dct['id']+'\' ri \'-1\' mt 1 \'1\' nt 43 ObID \'3B_PAVED\' px '+dct['px']+' py '+dct['py']+' node\n')
                  
            # Greenhouse nodes
            if any(self.rrmodel.greenhouse.gh_nodes):
                for _, dct in self.rrmodel.greenhouse.gh_nodes.items():
                    if float(dct['ar']) > 0.0:                    
                        f.write('NODE id \''+dct['id']+'\' nm \''+dct['id']+'\' ri \'-1\' mt 1 \'3\' nt 45 ObID \'3B_GREENHOUSE\' px '+dct['px']+' py '+dct['py']+' node\n')

            # Openwater nodes
            if any(self.rrmodel.openwater.ow_nodes):
                for _, dct in self.rrmodel.openwater.ow_nodes.items():
                    if float(dct['ar']) > 0.0:                    
                        f.write('NODE id \''+dct['id']+'\' nm \''+dct['id']+'\' ri \'-1\' mt 1 \'21\' nt 46 ObID \'OW_PRECIP\' px '+dct['px']+' py '+dct['py']+' node\n')
                      
            # Boundary_nodes
            if any(self.rrmodel.external_forcings.boundary_nodes):
                for _, dct in self.rrmodel.external_forcings.boundary_nodes.items():
                    f.write('NODE id \''+dct['id']+'\' nm \''+dct['id']+'\' ri \'-1\' mt 1 \'6\' nt 78 ObID \'SBK_SBK-3B-NODE\' px '+dct['px']+' py '+dct['py']+' node\n')
                    
        filepath = os.path.join(self.output_dir, '3B_LINK.TP')
        with open(filepath, 'w') as f:
          
            cnt = 0
            if any(self.rrmodel.unpaved.unp_nodes):                
                for _, dct in self.rrmodel.unpaved.unp_nodes.items():
                    if np.sum([float(d) for d in dct['ar'].split(' ')]) > 0.0:                    
                        cnt += 1
                        f.write('BRCH id \''+str(cnt)+'\' ri \'-1\' mt 1 \'0\' bt 17 ObID \'3B_LINK\' bn \''+dct['id']+'\' en \''+dct['boundary_node']+'\' brch\n')
            if any(self.rrmodel.paved.pav_nodes):                
                for _, dct in self.rrmodel.paved.pav_nodes.items():
                    if float(dct['ar']) > 0.0:                    
                        cnt += 1
                        f.write('BRCH id \''+str(cnt)+'\' ri \'-1\' mt 1 \'0\' bt 17 ObID \'3B_LINK\' bn \''+dct['id']+'\' en \''+dct['boundary_node']+'\' brch\n')
            if any(self.rrmodel.greenhouse.gh_nodes):                
                for _, dct in self.rrmodel.greenhouse.gh_nodes.items():
                    if float(dct['ar']) > 0.0:                    
                        cnt += 1
                        f.write('BRCH id \''+str(cnt)+'\' ri \'-1\' mt 1 \'0\' bt 17 ObID \'3B_LINK\' bn \''+dct['id']+'\' en \''+dct['boundary_node']+'\' brch\n')
            if any(self.rrmodel.openwater.ow_nodes):                
                for _, dct in self.rrmodel.openwater.ow_nodes.items():
                    if float(dct['ar']) > 0.0:                    
                        cnt += 1
                        f.write('BRCH id \''+str(cnt)+'\' ri \'-1\' mt 1 \'0\' bt 17 ObID \'3B_LINK\' bn \''+dct['id']+'\' en \''+dct['boundary_node']+'\' brch\n')

        # bound3b.3b        
        filepath = os.path.join(self.output_dir, 'Bound3B.3B')
        with open(filepath, 'w') as f:
            for _, dct in self.rrmodel.external_forcings.boundary_nodes.items():                
                f.write('BOUN id \''+dct['id']+'\' bl 2 \'0\' is 0 boun\n')
                
        # BoundaryConditions.bc
        filepath = os.path.join(self.output_dir, 'BoundaryConditions.bc')
        header = {'majorVersion':'1', 'minorVersion':'0', 'fileType':'boundConds'}
        with open(filepath, 'w') as f:                        
            self._write_dict(f, header, 'General','\n')            
            for _, dct in self.rrmodel.external_forcings.boundary_nodes.items():                
                temp = {"name":''+dct['id'], 'function':'constant','quantity':'water_level','unit':'m'} 
                self._write_dict(f,temp,'Boundary','    0\n\n')
                                   
    def write_unpaved(self):
        """
        Method to write all files associated with unpaved nodes: UNPAVED.3B, UNPAVED.ALF, UNPAVED.STO, UNPAVED.INF and UNPAVED.SEP. All files contain a definition for every node  because they may or may not be spatially distributed.

        Returns
        -------
        None.

        """
        if any(self.rrmodel.unpaved.unp_nodes): 
            filepath = os.path.join(self.output_dir, 'UNPAVED.3B')
            with open(filepath, 'w') as f:
                for _, dct in self.rrmodel.unpaved.unp_nodes.items():
                    if np.sum([float(d) for d in dct['ar'].split(' ')]) > 0.0:                    
                        f.write('UNPV id \''+dct['id']+'\' na '+dct['na']+' gw '+dct['ga']+' ar '+dct['ar']+' lv '+dct['lv']+' co '+dct['co']+' su '+dct['su']+' sd \'sto_'+dct['id']+'\' ic \'inf_'+dct['id']+'\' bt '+dct['bt']+' ed \''+dct['ed']+'\' sp \''+dct['sp']+'\' ig 0 '+dct['ig']+' mg '+dct['mg']+' gl '+dct['gl']+' is '+dct['is']+' ms \''+dct['ms']+'\' unpv\n')                    
                
            filepath = os.path.join(self.output_dir, 'UNPAVED.STO')
            with open(filepath, 'w') as f:            
                for _, dct in self.rrmodel.unpaved.unp_nodes.items():
                    if np.sum([float(d) for d in dct['ar'].split(' ')]) > 0.0:
                        f.write('STDF id \'sto_'+dct['id']+'\' nm \'sto_'+dct['id']+'\' ml '+dct['sd']+' il 0 stdf\n')
                
            filepath = os.path.join(self.output_dir, 'UNPAVED.INF')
            with open(filepath, 'w') as f:            
                for _, dct in self.rrmodel.unpaved.unp_nodes.items():
                    if np.sum([float(d) for d in dct['ar'].split(' ')]) > 0.0:
                        f.write('INFC id \'inf_'+dct['id']+'\' nm \'inf_'+dct['id']+'\' ic '+dct['ic']+' infc\n')
    
            filepath = os.path.join(self.output_dir, 'UNPAVED.ALF')
            with open(filepath, 'w') as f:            
                for _, dct in self.rrmodel.unpaved.ernst_defs.items():                    
                        f.write('ERNS id \''+dct['id']+'\' nm \''+dct['id']+'\' cvi '+dct['cvi']+' cvo 0 '+dct['cvo']+' cvs '+dct['cvs']+' lv '+dct['lv']+' erns\n')                    
                                                
            filepath = os.path.join(self.output_dir, 'UNPAVED.SEP')
            with open(filepath, 'w') as f:            
                for _, dct in self.rrmodel.unpaved.unp_nodes.items():                                            
                    if np.sum([float(d) for d in dct['ar'].split(' ')]) > 0.0:
                        f.write('SEEP id \''+dct['sp']+'\' nm \''+dct['sp']+'\' co 4 PDIN 1 1 \'365;00:00:00\' pdin ss 0\n')
                        f.write('TBLE\n')
                        for row in self.rrmodel.external_forcings.seepage[dct['sp']]['seepage'].iteritems():
                            f.write('\''+row[0].strftime('%Y/%m/%d;%H:%M:%S')+'\' '+f'{row[1]:.5f} <\n')
                        f.write('tble\nseep\n')              
            
    def write_paved(self):
        """
        Method to write all files associated with paved nodes: PAVED.3B, PAVED.STO and PAVED.DWA. The latter contains only one definition.

        Returns
        -------
        None.

        """
        if any(self.rrmodel.paved.pav_nodes):
            filepath = os.path.join(self.output_dir, 'PAVED.3B')
            with open(filepath, 'w') as f:          
                for _, dct in self.rrmodel.paved.pav_nodes.items():                    
                    if float(dct['ar']) > 0.0:                    
                        f.write('PAVE id \''+dct['id']+'\' ar '+dct['ar']+' lv '+dct['lv']+' sd \'sto_'+dct['id']+'\' ss 0 qc 0 '+dct['qc']+' 0 qo 1 0 ms \''+dct['ms']+'\' aaf 1 is '+dct['is']+' np '+dct['np']+' dw \'Def_DWA\' ro '+dct['ro']+' ru '+dct['ru']+' qh \'\' pave\n')
                                                                                                                                                                                                           
            filepath = os.path.join(self.output_dir, 'PAVED.STO')                       
            with open(filepath, 'w') as f:           
                for _, dct in self.rrmodel.paved.pav_nodes.items():                    
                    if float(dct['ar']) > 0.0:                                   
                        f.write('STDF id \'sto_'+dct['id']+'\' nm \'sto_'+dct['id']+'\' ms '+dct['strs']+' is 0 mr '+dct['sews']+' 0.0 ir 0.0 0.0 stdf\n')
                    
            filepath = os.path.join(self.output_dir, 'PAVED.DWA')             
            with open(filepath, 'w') as f:                           
                f.write('DWA id \'Def_DWA\' nm \'Def_DWA\' do 1 wc 0 wd 0 wh 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 dwa\n')
                    
    def write_greenhouse(self):
        """
        Method to write all files associated with greenhouse nodes: GREENHSE.3B, GREENHSE.RF and GREENHSE.SIL. The latter contains only one definition.

        Returns
        -------
        None.

        """
        if any(self.rrmodel.greenhouse.gh_nodes):
            filepath = os.path.join(self.output_dir, 'GREENHSE.3B')
            with open(filepath, 'w') as f:                    
                for _, dct in self.rrmodel.greenhouse.gh_nodes.items():    
                    if float(dct['ar']) > 0.0:                    
                        f.write('GRHS id \''+dct['id']+'\' na 10 ar 0 0 '+dct['ar']+' 0 0 0 0 0 0 0 sl '+dct['sl']+' as 0 si \'Def_silo\' sd \'sto_'+dct['id']+'\' ms \''+dct['ms']+'\' is '+dct['is']+' grhs\n')      
                       
            filepath = os.path.join(self.output_dir, 'GREENHSE.SIL')
            with open(filepath, 'w') as f:
                f.write('SILO id \'Def_silo\' nm \'Def_silo\' sc 0.0 pc 0.0 silo\n')
            
            filepath = os.path.join(self.output_dir, 'GREENHSE.RF')
            with open(filepath, 'w') as f:
                for _, dct in self.rrmodel.greenhouse.gh_nodes.items():    
                    if float(dct['ar']) > 0.0:                                    
                        f.write('STDF id \'sto_'+dct['id']+'\' nm \'sto_'+dct['id']+'\' mk '+dct['sd']+' ik 0 stdf\n')
    
    def write_openwater(self):
         """
         Method to write OPENWATE.3B file.

        Returns
        -------
        None.

        """
         # write openwater.3b
         if any(self.rrmodel.openwater.ow_nodes):
            filepath = os.path.join(self.output_dir, 'OPENWATE.3B')
            with open(filepath, 'w') as f:                    
                for _, dct in self.rrmodel.openwater.ow_nodes.items(): 
                    if float(dct['ar']) > 0.0:                    
                        f.write('OWRR id \''+dct['id']+'\' ar '+dct['ar']+' ms \''+dct['ms']+'\' owrr\n')                                 
         
    def write_meteo(self):      
        """
        Method to write meteofiles (DEFAULT.BUI and DEFAULT.EVP) based on values per catchment.

        Returns
        -------
        None.

        """            
         # precip         
        df = {}
        for ii,ms in enumerate( self.rrmodel.external_forcings.precip.items()):                 
            df[ms[0]] = pd.Series(ms[1]['precip'])
        filepath = os.path.join(self.output_dir, 'DEFAULT.BUI')
        f = open(filepath, 'w')
        f.write('*Name of this file: c:\Result\1058\DEFAULT.BUI\n')
        f.close()
        f = open(filepath,'a')
        f.write('*Date and time of construction: 00/00/2000 00:00:00.\n')
        f.write('1\n')
        f.write('*Aantal stations\n')
        f.write(str(len(self.rrmodel.external_forcings.precip))+'\n')
        f.write('*Namen van stations\n')
        for ms in self.rrmodel.external_forcings.precip.items():
            f.write('\''+ms[0]+'\''+'\n' )
        f.write('*Aantal gebeurtenissen (omdat het 1 bui betreft is dit altijd 1)\n')
        f.write('*en het aantal seconden per waarnemingstijdstap\n')
        f.write(f'1 {(ms[1]["precip"].index[1]-ms[1]["precip"].index[0]).total_seconds():.0f}\n')
        f.write('*Elke commentaarregel wordt begonnen met een * (asterisk).\n')
        f.write('*Eerste record bevat startdatum en -tijd, lengte van de gebeurtenis in dd hh mm ss\n')
        f.write('*Het format is: yyyymmdd:hhmmss:ddhhmmss')
        f.write('*Daarna voor elk station de neerslag in mm per tijdstap.\n')
        duration = (ms[1]['precip'].index[-1]-ms[1]['precip'].index[0]).total_seconds()
        f.write(ms[1]['precip'].index[0].strftime("%Y %#m %#d %#H %#M %#S ")+str(int(duration/86400.))+' '+str(int(np.remainder(duration,86400.)/3600.))+' 0 0\n')                        
        f.close()
        self._dict_to_df()
        self.precip_df.to_csv(filepath, header=False, index=False, sep=" ", float_format='%.3f', mode='a')
#%%
          
#%%%          
        # evaporation         
        filepath = os.path.join(self.output_dir, 'DEFAULT.EVP')
        f = open(filepath,'w')
        f.write('*Verdampingsfile\n')
        f.write('*Meteo data: evaporation intensity in mm/day\n')
        f.write('*First record: start date, data in mm/day\n')
        f.write('*Datum (year month day), verdamping (mm/dag) voor elk weerstation\n')
        f.write('*jaar maand dag verdamping[mm]\n')     
        f.close()
        table = list(self.rrmodel.external_forcings.evap.values())[0]['evap']          
        table.to_csv(filepath,float_format='%.3f', date_format='%#Y  %#m  %#d ', sep=" ", header=False ,quoting=csv.QUOTE_NONE, mode="a", escapechar=" ")         
               
    def _dict_to_df(self):        
        """
        Method to convert the precip dictionary to a dataframe
        Returns
        -------
        None.

        """
        df = {}
        for ii, stat in enumerate(self.rrmodel.external_forcings.precip.items()):              
            df[stat[0]] = pd.Series(stat[1]['precip'])        
        self.precip_df = pd.DataFrame(df)
    
    def _write_dict(self, f, dct, header, endline):
        """
        Write a dictionary to file, without transforming with mapping
        """
        # Write header
        f.write(f'[{header}]\n')
        for key, value in dct.items():
            f.write(f'    {key:20s} = {value:20s}\n')
        f.write(endline)
        
    def change_d3b_parameter(self, parameter, value):
        """
        Change parameter value in delft3b.ini file
        """
        with open(os.path.join(self.output_dir,'DELFT_3B.INI'), 'r') as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            if line.lower().startswith(parameter.lower()):
                items = re.split('=|#|$\s+', line)
                key, oldvalue = items[0], items[1]
                if not key.strip().lower() == parameter.lower():
                    continue
                lines[i] = line.replace('='+oldvalue, f'={value}'.ljust(len(oldvalue)+1))
                break

        with open(os.path.join(self.output_dir, 'DELFT_3B.INI'), 'w') as f:
            f.write(''.join(lines))
      
    def write_coupling(self):        
        """
        Method to write the files required for coupling DFM and RR.
        """
        # run.bat
        with open(os.path.join(self.output_dir, '../run.bat'),'w') as f:
                 f.write('@ echo off\n')
                 f.write('set OMP_NUM_THREADS=2\n')
                 f.write('call \"'+self.run_dimrpad+'\"\n')
                 f.write('pause\n')

        # coupling XML
        DFM_comp_name = 'CoupledTest1_DFM'
        RR_comp_name = 'CoupledTest1_RR'
        with open(os.path.join(self.output_dir, '../dimr_config.xml'),'w') as f:       
                 f.write('<?xml version="1.0" encoding="utf-8" standalone="yes"?>\n')
                 f.write('<dimrConfig xmlns="http://schemas.deltares.nl/dimr" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://schemas.deltares.nl/dimr http://content.oss.deltares.nl/schemas/dimr-1.2.xsd">\n')
                 f.write('\t<documentation>\n')
                 f.write('\t\t<fileVersion>1.2</fileVersion>\n')
                 f.write('\t\t<createdBy>D-HyDAMO in delft3dfmpy v.'+self.version['number']+'</createdBy>\n')
                 f.write("\t\t<creationDate>"+self.version['date']+'</creationDate>\n')
                 f.write('\t</documentation>\n\n')        
                 f.write('\t<control>\n')
                 f.write('\t\t<parallel>\n')
                 f.write('\t\t\t<startGroup>\n')
                 tstr = f'0 {self.rrmodel.d3b_parameters["Timestepsize"]:.0f} {(pd.Timestamp(self.rrmodel.d3b_parameters["EndTime"])-pd.Timestamp(self.rrmodel.d3b_parameters["StartTime"])).total_seconds():.0f}'
                 f.write('\t\t\t\t<time>'+tstr+'</time>\n')
                 f.write('\t\t\t\t<coupler name="flow2rr" />\n')
                 f.write('\t\t\t\t<start name="'+RR_comp_name+'" />\n')
                 f.write('\t\t\t\t<coupler name="rr2flow" />\n')
                 f.write('\t\t\t</startGroup>\n')
                 f.write('\t\t\t<start name="'+DFM_comp_name+'" />\n')
                 f.write('\t\t</parallel>\n')
                 f.write('\t</control>\n\n')                                      

                 f.write('\t<component name="'+DFM_comp_name+'">\n')
                 f.write('\t\t<library>dflowfm</library>\n')
                 f.write('\t\t<workingDir>fm</workingDir>\n')
                 f.write('\t\t<inputFile>'+self.name+'.mdu</inputFile>\n')
                 f.write('\t</component>\n')                 
                 f.write('\t<component name="'+RR_comp_name+'">\n')
                 f.write('\t\t<library>rr_dll</library>\n')
                 f.write('\t\t<workingDir>rr</workingDir>\n')
                 f.write('\t\t<inputFile>Sobek_3b.fnm</inputFile>\n')
                 f.write('\t</component>\n\n')
        
                 f.write('\t<coupler name="flow2rr">\n')
                 f.write('\t\t<sourceComponent>'+DFM_comp_name+'</sourceComponent>\n')
                 f.write('\t\t<targetComponent>'+RR_comp_name+'</targetComponent>\n')		     
                 for i in self.rrmodel.external_forcings.boundary_nodes.items():
                     f.write('\t\t\t<item>\n')                     
                     # FOr now use one observation ponit- until water levels can be read from other nodes
                     #f.write('\t\t\t\t<sourceName>observations/'+str(i[0])+'/water_level</sourceName>\n')
                     f.write('\t\t\t\t<sourceName>laterals/'+str(i[0])+'/water_level</sourceName>\n')
                     f.write('\t\t\t\t<targetName>catchments/'+str(i[0])+'/water_level</targetName>\n')
                     f.write('\t\t\t</item>\n')                                  
                 f.write('\t\t<logger>\n')
                 f.write('\t\t\t<workingDir>.</workingDir>\n')
                 f.write('\t\t\t<outputFile>dflowfm_to_rr.nc</outputFile>\n')
                 f.write('\t\t</logger>\n')
                 f.write('\t</coupler>\n\n')
                 
                 f.write('\t<coupler name="rr2flow">\n')
                 f.write('\t\t<sourceComponent>'+RR_comp_name+'</sourceComponent>\n')
                 f.write('\t\t<targetComponent>'+DFM_comp_name+'</targetComponent>\n')		         
                 for i in self.rrmodel.external_forcings.boundary_nodes.items():
                     f.write('\t\t\t<item>\n')                     
                     f.write('\t\t\t\t<sourceName>catchments/'+str(i[0])+'/water_discharge</sourceName>\n')
                     f.write('\t\t\t\t<targetName>laterals/'+str(i[0])+'/water_discharge</targetName>\n')
                     f.write('\t\t\t</item>\n')                                 

                 f.write('\t\t<logger>\n')
                 f.write('\t\t\t<workingDir>.</workingDir>\n')
                 f.write('\t\t\t<outputFile>rr_to_dflowfm.nc</outputFile>\n')
                 f.write('\t\t</logger>\n')                 
                 f.write('\t</coupler>\n')
                 f.write('</dimrConfig>\n')
        
    


#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['delft3dfmpy']

package_data = \
{'': ['*'],
 'delft3dfmpy': ['converters/*',
                 'core/*',
                 'datamodels/*',
                 'io/*',
                 'io\\resources/*',
                 'io\\resources\\RRmodel/*',
                 'lib/*']}

install_requires = \
['geopandas',
 'rasterio',
 'imod',
 'numpy',
 'scipy',
 'pandas',
 'matplotlib',
 'pillow',
 'rasterstats',
 'tqdm',
 'netcdf4',
 'descartes',
 'contextily',
 'hkvsobekpy']

extras_require = \
{'dev': ['pytest',
         'pytest-cov',
         'black',
         'twine',
         'bumpversion',
         'watchdog',
         'coverage',
         'Sphinx']}

setup(name='delft3dfmpy',
      version='1.1.1',
      description='Top-level package for delft3dfmpy.',
      author='Deltares - HKV',
      author_email=None,
      url='https://github.com/openearth/delft3dfmpy',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )




import logging
import os
from ctypes import CDLL, byref, c_char, c_int, pointer
from itertools import combinations

import geopandas as gpd
import numpy as np
from scipy.interpolate import LinearNDInterpolator
from scipy.spatial import KDTree, Voronoi
from shapely.geometry import (
    LineString, MultiLineString, MultiPolygon, Point, Polygon, box)
from shapely.ops import unary_union
from shapely.prepared import prep

from delft3dfmpy.core import checks, geometry
from delft3dfmpy.datamodels.cstructures import meshgeom, meshgeomdim
from delft3dfmpy.io import gridio

logger = logging.getLogger(__name__)


class Mesh2D:

    def __init__(self):
        self.fill_value_z = -999.0
        self.missing_z_value = None

        self.meshgeomdim = meshgeomdim(pointer(c_char()), 2, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0)
        self.meshgeom = meshgeom(self.meshgeomdim)

    def clip_nodes(self, xnodes, ynodes, edge_nodes, polygon, keep_outside=False):
        """
        Method that returns a set of nodes that is clipped within a given polygon.

        Parameters
        ----------
        xnodes : 
            X-coordinates of mesh nodes
        ynodes : 
            Y-coordinates of mesh nodes
        edge_nodes : 
            node ids of edges. These are assumed in 1-based numbering
        polygon : Polygon, Multipolygon or list of
        
        """

        # Find nodes within one of the polygons
        logger.info('Selecting nodes within polygon.')
        index = np.zeros(len(xnodes), dtype=bool)
        for poly in geometry.as_polygon_list(polygon):
            index = index | geometry.points_in_polygon(np.c_[xnodes, ynodes], poly)
        if keep_outside:
            index = ~index

        # Filter edge nodes, keep the ones that are within a polygon with both points
        nodeids_in_polygon = np.where(index)[0] + 1
        edge_nodes = edge_nodes[np.isin(edge_nodes, nodeids_in_polygon).all(axis=1)]

        # Remove edges that intersect the clipping polygons. This is needed for when the clipping polygon is narrow,
        # So it crosses an edge without containing one of the edge nodes
        logger.info('Creating LineString for each edge.')
        edges = [LineString(line) for line in np.c_[xnodes, ynodes][edge_nodes - 1]]
        
        logger.info('Checking for intersections.')
        isect = np.zeros(len(edge_nodes), dtype=bool)
        for poly in geometry.as_polygon_list(polygon):
            prepped = prep(poly.exterior)
            isect = isect | np.array([prepped.intersects(e) for e in edges])
            for interior in poly.interiors:
                prepped = prep(interior)
                isect = isect | np.array([prepped.intersects(e) for e in edges])
        edge_nodes = edge_nodes[~isect]

        # Remove edges that have a 1 count, until none are left
        logger.info('Remove edges with only a single node within the clip area.')
        unique, count = np.unique(edge_nodes, return_counts=True)
        while any(count == 1):
            count1nodes = unique[count == 1]
            edge_nodes = edge_nodes[~np.isin(edge_nodes, count1nodes).any(axis=1)]
            unique, count = np.unique(edge_nodes, return_counts=True)

        # Select remaining nodes
        xnodes = xnodes[unique - 1]
        ynodes = ynodes[unique - 1]

        # Make mapping for new id's
        new_id_mapping = {old_id: new_id + 1 for new_id, old_id in enumerate(unique)}
        edge_nodes = np.reshape([new_id_mapping[old_id] for old_id in edge_nodes.ravel()], edge_nodes.shape)

        return xnodes, ynodes, edge_nodes

    def clip_mesh_by_polygon(self, polygon, outside=True):
        """
        Removes part of the existing mesh within the given polygon.

        Parameters
        ----------
        polygon : list, Polygon, MultiPolygon
            A polygon, multi-polygon or list of multiple polygons or multipolygons.
            The faces within the polygons are removed.
        """
        logger.info('Clipping 2D mesh by polygon.')

        # Check if the input arguments is indeed some kind of polygon
        checks.check_argument(polygon, 'polygon', (list, Polygon, MultiPolygon))
        
        # Get xnodes, ynodes and edge_nodes
        xnodes = self.meshgeom.get_values('nodex', as_array=True)
        ynodes = self.meshgeom.get_values('nodey', as_array=True)
        edge_nodes = self.meshgeom.get_values('edge_nodes', as_array=True)

        # Clip
        logger.info('Clipping nodes.')
        xnodes, ynodes, edge_nodes = self.clip_nodes(xnodes, ynodes, edge_nodes, polygon, keep_outside=outside)
        # edge_nodes = np.sort(edge_nodes, axis=1)

        # Update dimensions
        self.meshgeomdim.numnode = len(xnodes)
        self.meshgeomdim.numedge = len(edge_nodes)

        # Add nodes and links
        self.meshgeom.set_values('nodex', xnodes)
        self.meshgeom.set_values('nodey', ynodes)
        self.meshgeom.set_values('edge_nodes', np.ravel(edge_nodes).tolist())

        # Determine what the cells are
        logger.info('Finding cell faces within clipped nodes.')
        self._find_cells(self.meshgeom)

        # Clean nodes, this function deletes nodes based on no longer existing cells
        face_nodes = self.meshgeom.get_values('face_nodes', as_array=True)
        logger.info('Removing incomplete faces.')
        cleaned = self.clean_nodes(xnodes, ynodes, edge_nodes, face_nodes)
       
        # If cleaning leaves no whole cells, return None
        if cleaned is None:
            raise ValueError('Clipping the grid does not leave any nodes.')
        # Else, get the arguments from the returned tuple
        else:
            xnodes, ynodes, edge_nodes, face_nodes = cleaned

        logger.info('Update mesh with clipped nodes, edges and faces.')
        # Update dimensions
        self.meshgeomdim.numnode = len(xnodes)
        self.meshgeomdim.numedge = len(edge_nodes)

        # Add nodes and links
        self.meshgeom.set_values('nodex', xnodes)
        self.meshgeom.set_values('nodey', ynodes)
        self.meshgeom.set_values('edge_nodes', np.ravel(edge_nodes).tolist())
        self.meshgeom.set_values('face_nodes', np.ravel(face_nodes).tolist())

    def clean_nodes(self, xnodes, ynodes, edge_nodes, face_nodes):
        """
        Method to clean the nodes. Edges that do not form a cell are deleted.
        Alternative method based on checking tuple pairs
        """

        # Select nodes the occur in any face
        node_selection = np.unique(face_nodes)
        node_selection = node_selection[node_selection!=-999]
        # If no nodes (because no faces), return None
        if not node_selection.any():
            return None
        
        # Remove all edges that are not in this selection
        edge_nodes = edge_nodes[np.isin(edge_nodes, node_selection).all(axis=1)]
                
        # Remove all segments that are not part of any face
        # Direction is irrelevant, sort to compare in ascending order
        # Convert to a contiguous array for faster comparison
        sorted_edges = np.ascontiguousarray(np.sort(edge_nodes, axis=1)).view(np.dtype((np.void, edge_nodes.dtype.itemsize * edge_nodes.shape[1])))
        idx = np.zeros(len(sorted_edges), dtype=bool)

        # Get combinations of face nodes that can form an edge
        combs = list(combinations(range(len(face_nodes[0])), 2))
        for comb in combs:
            # Get the first, second, third, etc. possible edge of every face
            arr = face_nodes[:, comb]
            # Remove nodata (so for example the non-existing fourth edge of a triangular face)
            arr = arr[(arr != -999).all(axis=1)]
            arr = np.sort(arr, axis=1)
            # Convert to a contiguous array for faster comparison
            sorted_face_edges = np.ascontiguousarray(arr).view(np.dtype((np.void, arr.dtype.itemsize * arr.shape[1])))
            # Check if face edges are in edges
            idx |= np.isin(sorted_edges, sorted_face_edges)[:, 0]

        # Select the nodes that are present in faces
        edge_nodes = edge_nodes[idx]

        xnodes = xnodes[node_selection - 1]
        ynodes = ynodes[node_selection - 1]

            # Make mapping for new id's
        new_id_mapping = {old_id: new_id + 1 for new_id, old_id in enumerate(node_selection)}    
            
        edge_nodes = np.reshape([-999 if old_id == -999 else new_id_mapping[old_id] for old_id in edge_nodes.ravel()], edge_nodes.shape)
        face_nodes = np.reshape([-999 if old_id == -999 else new_id_mapping[old_id] for old_id in face_nodes.ravel()], face_nodes.shape)
        
        return xnodes, ynodes, edge_nodes, face_nodes

    @staticmethod
    def _find_cells(geometries, maxnumfacenodes=None):
        """
        Determine what the cells are in a grid.
        """

        dimensions = geometries.meshgeomdim

        wrapperGridgeom = CDLL(os.path.join(os.path.dirname(__file__), '..', 'lib', 'gridgeom.dll'))
        ierr = wrapperGridgeom.ggeo_deallocate()
        assert ierr == 0

        start_index = c_int(1)

        meshdimout = meshgeomdim()
        meshout = meshgeom(meshdimout)

        ierr = wrapperGridgeom.ggeo_find_cells(
            byref(dimensions),
            byref(geometries),
            byref(meshdimout),
            byref(meshout),
            byref(start_index)
        )
        assert ierr == 0

        dimensions.numface = meshdimout.numface
        if maxnumfacenodes is None:
            dimensions.maxnumfacenodes = meshdimout.maxnumfacenodes
        else:
            dimensions.maxnumfacenodes = maxnumfacenodes

        # Allocate
        for mesh in [geometries, meshout]:
            for var in ['facex', 'facey', 'face_nodes']:
                mesh.allocate(var)

        ierr = wrapperGridgeom.ggeo_find_cells(
            byref(dimensions),
            byref(geometries),
            byref(meshdimout),
            byref(meshout),
            byref(start_index)
        )
        assert ierr == 0

        # Copy content to self meshgeom
        for var in ['facex', 'facey', 'face_nodes']:
            # Create array of allocated size. In case of deviating maxnumfacenodes
            if (maxnumfacenodes is not None) and (var == 'face_nodes'):
                # Create an empty integer. Empty values should only be present in the integer arrays
                arr = np.full(geometries.get_dimensions(var), -999, dtype=int)
                # Get array
                arr[:, :meshout.meshgeomdim.maxnumfacenodes] = meshout.get_values(var, as_array=True)
                # Set array
                geometries.set_values(var, arr.ravel())
            else:
                # Set array
                geometries.set_values(var, meshout.get_values(var))
    

        ierr = wrapperGridgeom.ggeo_deallocate()
        
    def altitude_constant(self, constant, where='face'):

        zvalues = np.ones(getattr(self.meshgeomdim, f'num{where}')) * constant
        self.meshgeom.set_values(f'{where}z', zvalues)

    def altitude_from_raster(self, rasterpath, where='face', stat='mean', missing='default'):
        """
        Method to determine level of nodes

        This function works faster for large amounts of cells, since it does not
        draw polygons but checks for nearest neighbours (voronoi) based
        on interpolation.

        Note that the raster is not clipped. Any values outside the bounds are
        also taken into account.
        """

        # Select points on faces or nodes
        logger.info('Creating GeoDataFrame of cell faces.')
        xy = np.c_[self.meshgeom.get_values(f'{where}x'), self.meshgeom.get_values(f'{where}y')]
        cells = self.meshgeom.get_faces()
        facedata = gpd.GeoDataFrame(geometry=[Polygon(cell) for cell in cells])   

        if where == 'face':
            # Get raster statistics
            facedata.index = np.arange(len(xy), dtype=np.uint32) + 1
            # facedata.to_file('test_face.shp')
            facedata['crds'] = [cell for cell in cells]
        
            df = geometry.raster_stats_fine_cells(rasterpath, facedata, stats=[stat])
            # Get z values
            zvalues = df[stat].values

        elif where == 'node':

            logger.info('Generating voronoi polygons around cell centers for determining raster statistics.')
            
            # Creat voronoi polygon
            # Add border to limit polygons
            border = box(xy[:, 0].min(), xy[:, 1].min(), xy[:, 0].max(), xy[:, 1].max()).buffer(1000).exterior
            borderpts = [border.interpolate(dist).coords[0] for dist in np.linspace(0, border.length, max(20, border.length//100))]
            vor = Voronoi(points=xy.tolist()+borderpts)
            clippoly = facedata.unary_union
            # Get lines
            lines = []
            for poly in geometry.as_polygon_list(clippoly):
                lines.append(poly.exterior)
                lines.extend([line for line in poly.interiors])
            linesprep = prep(MultiLineString(lines))
            clipprep = prep(clippoly)

            # Collect polygons
            data = []
            for (pr, (i, pt)) in zip(vor.point_region, enumerate(xy)):
                region = vor.regions[pr]
                if pr == -1:
                    break
                while -1 in region:
                    region.remove(-1)
                if len(region) < 3:
                    continue
                crds = vor.vertices[region]
                if clipprep.intersects(Point(pt)):
                    poly = Polygon(crds)
                    if linesprep.intersects(poly):
                        poly = poly.intersection(clippoly)
                        if isinstance(poly, MultiPolygon):
                            poly = poly.buffer(0.001)
                        if isinstance(poly, MultiPolygon):
                            logger.warning('Got multipolygon when clipping voronoi polygon. Only adding coordinates for largest of the polygons.')
                            poly = poly[np.argmax([p.area for p in as_polygon_list(poly)])]
                        crds = np.vstack(poly.exterior.coords[:])
                    data.append({'geometry': poly, 'crds': crds})
                    
            # Limit to model extend
            facedata = gpd.GeoDataFrame(data)
            facedata.index=np.arange(len(xy), dtype=np.uint32) + 1
            # facedata.drop('crds', axis=1).to_file('test_node.shp')
            # Get raster statistics
            df = geometry.raster_stats_fine_cells(rasterpath, facedata, stats=[stat])
            # Get z values
            zvalues = df[stat].values

        else:
            raise NotImplementedError()


        # If there are no NaN's, return the answer
        isnan = np.isnan(zvalues)
        if not isnan.any():
            # Set values to mesh geometry
            self.meshgeom.set_values(f'{where}z', zvalues)
            return None

        # If interpolation or missing, but all are NaN, raise error.
        elif isnan.all() and missing in ['nearest', 'interpolation']:
            raise ValueError('Only NaN values found, interpolation or nearest not possible.')
        
        # Fill missing values
        if missing == 'default':
            # With default value
            zvalues[isnan] = self.fill_value_z

        elif isinstance(missing, (float, int)):
            # With a given number
            zvalues[isnan] = missing

        elif missing == 'nearest':
            # By looking for the nearest value in the grid
            # Create a KDTree of the known points
            tree = KDTree(data=xy[~isnan])
            idx = tree.query(x=xy[isnan])[1]
            zvalues[isnan] = zvalues[~isnan][idx]

        elif missing == 'interpolation':
            # By interpolating
            isnan = np.isnan(zvalues)
            interp = LinearNDInterpolator(xy[~isnan], zvalues[~isnan], fill_value=self.fill_value_z)
            zvalues[isnan] = interp(xy[isnan])

        else:
            raise ValueError(f'{missing} not recognized. Choose \'default\', \'nearest\', \'interpolation\' or a number with which to fill the missing data.')

        # Set values to mesh geometry
        self.meshgeom.set_values(f'{where}z', zvalues)

    def set_missing_z_value(self, value):
        self.missing_z_value = value

    def geom_from_netcdf(self, file, only2d=False):
        gridio.from_netcdf_old(self.meshgeom, file, only2d=only2d)
        self._find_cells(self.meshgeom)

    def faces_to_centroid(self):
        """
        The find_cells function does not always return face coordinates
        that are all accepted by the interactor. The coordinates
        are placed on the circumcenters, and placed on the face border
        in case of a coordinate that is outside the cells.
        
        This function shifts the face coordinates towards the centroid
        (center of gravity) of the cell.
        """
        # Calculate centroids
        centroids = np.vstack([cell.mean(axis=0) for cell in self.meshgeom.get_faces()]).T

        # Set values back to geometry
        self.meshgeom.set_values('facex', centroids[0])
        self.meshgeom.set_values('facey', centroids[1])
                               
class Rectangular(Mesh2D):

    def __init__(self):
        Mesh2D.__init__(self)
        self.meshgeomdim.maxnumfacenodes = 4
        self.rotated = False

    def generate_grid(self, x0, y0, dx, dy, ncols, nrows, clipgeo=None, rotation=0):
        """
        Generate rectangular grid based on the origin (x0, y0) the cell sizes (dx, dy),
        the number of columns and rows (ncols, nrows) and a rotation in degrees (default=0)
        A geometry (clipgeo) can be given to clip the grid.
        """

        # Generate x and y spacing
        x = np.linspace(x0, x0 + dx * (ncols), ncols+1)
        y = np.linspace(y0, y0 + dy * (nrows), nrows+1)

        # Get nodes
        xnodes, ynodes = np.meshgrid(x, y)

        # Rotate
        if rotation != 0:
            self.rotated = True
            xnodes, ynodes = geometry.rotate_coordinates((x0, y0), np.radians(rotation), xnodes, ynodes)

        # Get nodes as list
        nodes = {crd: i+1 for i, crd in enumerate(zip(xnodes.ravel(), ynodes.ravel()))}

        # Create segments
        segments = list(zip(
            zip(*np.c_[xnodes[:, :-1].ravel(), ynodes[:, :-1].ravel()].T),
            zip(*np.c_[xnodes[:, 1:].ravel(), ynodes[:, 1:].ravel()].T)
        ))
        segments += list(zip(
            zip(*np.c_[xnodes[1:, :].ravel(), ynodes[1:, :].ravel()].T),
            zip(*np.c_[xnodes[:-1, :].ravel(), ynodes[:-1, :].ravel()].T)
        ))

        # Get links
        edge_nodes = np.asarray([(nodes[s[0]], nodes[s[1]]) for s in segments])

        # Rvael the nodes
        xnodes = xnodes.ravel()
        ynodes = ynodes.ravel()

        # Add to mesh
        dimensions = meshgeomdim()
        dimensions.dim = 2
        geometries = meshgeom(dimensions)

        # Clip
        if clipgeo is not None:
            xnodes, ynodes, edge_nodes = self.clip_nodes(xnodes, ynodes, edge_nodes, clipgeo)

        # Update dimensions
        dimensions.numnode = len(xnodes)
        dimensions.numedge = len(edge_nodes)

        # Add nodes and links
        geometries.set_values('nodex', xnodes)
        geometries.set_values('nodey', ynodes)
        geometries.set_values('edge_nodes', np.ravel(edge_nodes).tolist())

        # Determine what the cells are
        self._find_cells(geometries)

        # Clip
        if clipgeo is not None:
            # Clean nodes, this function deletes nodes based on no longer existing cells
            face_nodes = geometries.get_values('face_nodes', as_array=True)
            cleaned = self.clean_nodes(xnodes, ynodes, edge_nodes, face_nodes)
            # If cleaning leaves no whole cells, return None
            if cleaned is None:
                return None
            # Else, get the arguments from the returned tuple
            else:
                xnodes, ynodes, edge_nodes, face_nodes = cleaned

        # Update dimensions
        dimensions.numnode = len(xnodes)
        dimensions.numedge = len(edge_nodes)
        dimensions.maxnumfacenodes = 4

        # Add nodes and links
        geometries.set_values('nodex', xnodes)
        geometries.set_values('nodey', ynodes)
        geometries.set_values('edge_nodes', np.ravel(edge_nodes).tolist())
        geometries.set_values('face_nodes', np.ravel(face_nodes).tolist())

        # Add to mesh
        self.meshgeom.add_from_other(geometries)

    def generate_within_polygon(self, polygon, cellsize, rotation=0):
        """
        Function to generate a grid within a polygon. It uses the function
        'generate_grid' but automatically detects the extent.

        Parameters
        ----------
        polygon : (list of) shapely.geometry.Polygon or a shapely.geometry.MultiPolygon
        """

        checks.check_argument(polygon, 'polygon', (list, Polygon, MultiPolygon))

        polygons = geometry.as_polygon_list(polygon)
        
        logger.info(f'Generating grid with cellsize {cellsize} m and rotation {rotation} degrees.')
        convex = MultiPolygon(polygons).convex_hull
        
        # In case of a rotation, extend the grid far enough to make sure
        if rotation != 0:
            # Find a box that contains the whole polygon
            (x0, y0), xsize, ysize = geometry.minimum_bounds_fixed_rotation(convex, rotation)
        else:
            xsize = convex.bounds[2] - convex.bounds[0]
            ysize= convex.bounds[3] - convex.bounds[1]
            x0, y0 = convex.bounds[0], convex.bounds[1]

        self.generate_grid(
            x0=x0,
            y0=y0,
            dx=cellsize,
            dy=cellsize,
            ncols=int(xsize / cellsize) + 1,
            nrows=int(ysize / cellsize) + 1,
            clipgeo=polygon,
            rotation=rotation
        )

    def refine(self, polygon, level, cellsize, debug=False, dflowfm_path=None):
        """
        Parameters
        ----------
        polygon : list of tuples
            Polygon in which to refine
        level : int
            Number of times to refine
        """

        if self.rotated:
            raise NotImplementedError('Mesh refinement does not work for rotation grids.')
        
        checks.check_argument(polygon, 'polygon', (list, Polygon, MultiPolygon))

        # Save network as grid
        temppath = 'tempgrid_net.nc'
        gridio.to_netcdf_old(self.meshgeom, temppath)

        if isinstance(level, int):
            level = [level]
            polygon = [polygon]

        factor = 2 ** max(level)
        stepsize = cellsize / factor
        dxmin = stepsize
        dtmax = stepsize / (9.81) ** 0.5


        xnodes = self.meshgeom.get_values('nodex')
        ynodes = self.meshgeom.get_values('nodey')
        x0 = min(xnodes)
        ncols = int((max(xnodes) - x0) / cellsize)
        y0 = min(ynodes)
        nrows = int((max(ynodes) - y0) / cellsize)

        arr = np.ones((nrows * factor, ncols * factor)) * 4**max(level)

        for poly, lvl in zip(polygon, level):
            mask = geometry.geometry_to_mask(poly, (x0, y0), stepsize, shape=(nrows * factor, ncols * factor))
            arr[mask] = 4 ** (max(level) - lvl)

        # Create ascii grid based on polygon
        header = (
            'nCols        {0:d}\n'
            'nRows        {1:d}\n'
            'xllCorner    {2:.6f}\n'
            'yllCorner    {3:.6f}\n'
            'CellSize     {4:.6f}\n'
            'nodata_value 0\n'
        ).format(ncols * factor, nrows * factor, x0, y0 + stepsize, stepsize)

        values = '\n'.join(' '.join('%d' %x for x in y) for y in arr.squeeze())

        with open('temp_ascgrid.asc', 'w') as f:
            f.write(header + values)

        # Get dflow fm on system path
        if dflowfm_path is not None:
            if not os.path.exists(os.path.join(dflowfm_path, 'dflowfm.exe')):
                raise OSError('dflowfm.exe not found on given path')
            dflowfm_path = os.path.join(dflowfm_path, 'dflowfm.exe')
        else:
            dflowfm_path = 'dflowfm.exe'

        # Construct statement
        statement = f'"{dflowfm_path}" --refine:hmin={dxmin}:dtmax={dtmax}:connect=1:outsidecell=1 {temppath} temp_ascgrid.asc'
        if debug:
            statement += ' > log.txt'
        # Execute statement
        os.system(statement)

        if not debug:
            os.remove('temp_ascgrid.asc')

        # Read geometry and dimensions again
        if not os.path.exists('out_net.nc'):
            raise OSError(
                (f'Could not find "out_net.nc" in directory "{os.getcwd()}". '
                 f'The file should have been generated with the statement: "{statement}". '
                 'You can test this manually by running this statement from a command prompt launched from the directory named in this message.'))
        gridio.from_netcdf_old(self.meshgeom, 'out_net.nc')
        
        # Find cells
        self._find_cells(self.meshgeom)
    
        os.remove('out_net.nc')
        os.remove(temppath)
